#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from compose_x_common.compose_x_common import keyisset
from troposphere import AWS_ACCOUNT_ID, AWS_PARTITION, GetAtt, If, Ref, Sub
from troposphere.kms import Alias, Key

from ecs_composex.common import LOG, build_template
from ecs_composex.common.cfn_conditions import USE_STACK_NAME_CON_T
from ecs_composex.common.cfn_params import ROOT_STACK_NAME
from ecs_composex.common.compose_resources import XResource, set_resources
from ecs_composex.common.stacks import ComposeXStack
from ecs_composex.kms import metadata
from ecs_composex.kms.kms_params import KMS_KEY_ARN, KMS_KEY_ID, MOD_KEY, RES_KEY
from ecs_composex.kms.kms_perms import get_access_types
from ecs_composex.kms.kms_template import create_kms_template
from ecs_composex.resources_import import import_record_properties


def define_default_key_policy():
    """
    Function to return the default KMS management policy allowing root account access.
    :return: policy
    :rtype: dict
    """
    policy = {
        "Version": "2012-10-17",
        "Id": "auto-secretsmanager-1",
        "Statement": [
            {
                "Sid": "Allow direct access to key metadata to the account",
                "Effect": "Allow",
                "Principal": {
                    "AWS": Sub(
                        f"arn:${{{AWS_PARTITION}}}:iam::${{{AWS_ACCOUNT_ID}}}:root"
                    )
                },
                "Action": ["kms:*"],
                "Resource": "*",
                "Condition": {
                    "StringEquals": {"kms:CallerAccount": Ref(AWS_ACCOUNT_ID)}
                },
            }
        ],
    }
    return policy


class KmsKey(XResource):
    """
    Class to represent a KMS Key
    """

    policies_scaffolds = get_access_types()

    def init_outputs(self):
        self.output_properties = {
            KMS_KEY_ID: (self.logical_name, self.cfn_resource, Ref, None),
            KMS_KEY_ARN: (
                f"{self.logical_name}{KMS_KEY_ARN.return_value}",
                self.cfn_resource,
                GetAtt,
                KMS_KEY_ARN.return_value,
            ),
        }

    def define_kms_key(self):
        """
        Method to set the KMS Key
        """
        if not self.properties:
            props = {
                "Description": Sub(
                    f"{self.name} created in ${{{ROOT_STACK_NAME.title}}}"
                ),
                "Enabled": True,
                "EnableKeyRotation": True,
                "KeyUsage": "ENCRYPT_DECRYPT",
                "PendingWindowInDays": 7,
            }
        else:
            props = import_record_properties(self.properties, Key)
        if not keyisset("KeyPolicy", props):
            props.update({"KeyPolicy": define_default_key_policy()})
        props.update({"Metadata": metadata})
        LOG.debug(props)
        self.cfn_resource = Key(self.logical_name, **props)

    def handle_key_settings(self, template):
        """
        Method to add to the template for additional KMS key related resources.

        :param troposphere.Template template:
        """
        if self.settings and keyisset("Alias", self.settings):
            alias_name = self.settings["Alias"]
            if not (alias_name.startswith("alias/") or alias_name.startswith("aws")):
                alias_name = If(
                    USE_STACK_NAME_CON_T,
                    Sub(f"alias/${{AWS::StackName}}/{alias_name}"),
                    Sub(f"alias/${{{ROOT_STACK_NAME.title}}}/{alias_name}"),
                )
            elif alias_name.startswith("alias/aws") or alias_name.startswith("aws"):
                raise ValueError(f"Alias {alias_name} cannot start with alias/aws.")
            Alias(
                f"{self.logical_name}Alias",
                template=template,
                AliasName=alias_name,
                TargetKeyId=Ref(self.cfn_resource),
                Metadata=metadata,
            )


class XStack(ComposeXStack):
    """
    Class for KMS Root stack
    """

    def __init__(self, title, settings, **kwargs):
        set_resources(settings, KmsKey, RES_KEY, MOD_KEY)
        new_keys = [
            key
            for key in settings.compose_content[RES_KEY].values()
            if not key.lookup and not key.use
        ]
        if new_keys:
            stack_template = build_template("Root template for KMS")
            super().__init__(title, stack_template, **kwargs)
            create_kms_template(stack_template, new_keys, self)
        else:
            self.is_void = True
        for resource in settings.compose_content[RES_KEY].values():
            resource.stack = self



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

import re
from os import path

from ecs_composex.common.cfn_params import Parameter
from ecs_composex.common.ecs_composex import X_KEY
from ecs_composex.vpc.vpc_params import SG_ID_TYPE

MOD_KEY = path.basename(path.dirname(path.abspath(__file__)))
RES_KEY = f"{X_KEY}{MOD_KEY}"


FS_REGEXP = re.compile(
    r"^(arn:aws[-a-z]*:elasticfilesystem:[0-9a-z-:]+:file-system/fs-[0-9a-f]{8,40}|fs-[0-9a-f]{8,40})$"
)


FS_ID_T = "FilesystemId"
FS_ID = Parameter(
    FS_ID_T,
    Type="String",
    AllowedPattern=FS_REGEXP.pattern,
)

FS_PORT_T = "FilesystemPort"
FS_PORT = Parameter(
    FS_PORT_T, Type="Number", MinValue=1, MaxValue=((2 ** 16) - 1), Default=2049
)

FS_ARN_T = "FilesystemArn"
FS_ARN = Parameter(FS_ARN_T, return_value="Arn", Type="String")

FS_AS_ID_T = "EfsAccessPointId"
FS_AS_ID = Parameter(FS_AS_ID_T, Type="String")

FS_MNT_PT_SG_ID_T = "FilesystemMountPointSgId"
FS_MNT_PT_SG_ID = Parameter(FS_MNT_PT_SG_ID_T, return_value="GroupId", Type=SG_ID_TYPE)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to scan and find the DB and Secret for Lookup of x-rds
"""

import re

from compose_x_common.compose_x_common import keyisset

from ecs_composex.common import LOG
from ecs_composex.common.aws import (
    define_lookup_role_from_info,
    find_aws_resource_arn_from_tags_api,
)
from ecs_composex.elasticache import elasticache_params


def get_cluster_config(resource, cluster_name, session):
    client = session.client("elasticache")
    try:
        cluster_r = client.describe_cache_clusters(
            CacheClusterId=cluster_name,
            ShowCacheClustersNotInReplicationGroups=True,
            ShowCacheNodeInfo=True,
        )
        cluster = cluster_r["CacheClusters"][0]
        if cluster["Engine"] == "memcached":
            resource.port_attr = elasticache_params.CLUSTER_MEMCACHED_PORT
            return {
                elasticache_params.CLUSTER_MEMCACHED_ADDRESS.title: cluster[
                    "ConfigurationEndpoint"
                ]["Address"],
                elasticache_params.CLUSTER_MEMCACHED_PORT.title: cluster[
                    "ConfigurationEndpoint"
                ]["Port"],
                elasticache_params.CLUSTER_SG.title: [
                    cluster["SecurityGroups"][0]["SecurityGroupId"]
                ],
            }
        elif cluster["Engine"] == "redis":
            if keyisset("ReplicationGroupId", cluster):
                raise LookupError(
                    "The Cluster identified is part of a replication group."
                )
            resource.port_attr = elasticache_params.CLUSTER_REDIS_PORT
            return {
                elasticache_params.CLUSTER_REDIS_PORT.title: cluster["CacheNodes"][0][
                    "Endpoint"
                ]["Port"],
                elasticache_params.CLUSTER_REDIS_ADDRESS.title: cluster["CacheNodes"][
                    0
                ]["Endpoint"]["Address"],
                elasticache_params.CLUSTER_SG.title: [
                    cluster["SecurityGroups"][0]["SecurityGroupId"]
                ],
            }
    except client.exceptions.CacheClusterNotFoundFault:
        LOG.error(f"Could not find the configurations for cluster {cluster_name}")


def get_replica_group_config(resource, cluster_name, session):
    client = session.client("elasticache")
    try:
        cluster_r = client.describe_replication_groups(ReplicationGroupId=cluster_name)
        cluster = cluster_r["ReplicationGroups"][0]
        node_r = client.describe_cache_clusters(
            CacheClusterId=cluster["MemberClusters"][0]
        )
        sg_id = node_r["CacheClusters"][0]["SecurityGroups"][0]["SecurityGroupId"]
        resource.port_attr = elasticache_params.REPLICA_PRIMARY_PORT
        return {
            elasticache_params.REPLICA_PRIMARY_ADDRESS.title: cluster["NodeGroups"][0][
                "PrimaryEndpoint"
            ]["Address"],
            elasticache_params.REPLICA_PRIMARY_PORT.title: cluster["NodeGroups"][0][
                "PrimaryEndpoint"
            ]["Port"],
            elasticache_params.REPLICA_READ_ENDPOINT_ADDRESSES.title: [
                cluster["NodeGroups"][0]["ReaderEndpoint"]["Address"]
            ],
            elasticache_params.REPLICA_READ_ENDPOINT_PORTS.title: [
                cluster["NodeGroups"][0]["ReaderEndpoint"]["Port"]
            ],
            elasticache_params.CLUSTER_SG.title: [sg_id],
        }
    except client.exceptions.ReplicationGroupNotFoundFault as error:
        LOG.error(f"Could not fetch information about {cluster_name}")
        LOG.error(error)
        return None


def return_cluster_config(resource, cluster_arn, session):
    """
    Function to retrieve the DB information we need for services integration

    :param ecs_composex.elasticache.elasticache_stack.CacheCluster resource:
    :param cluster_arn:
    :param session:
    :type cluster_arn: str
    :type session: boto3.session.Session
    :return: the DB details
    """

    if isinstance(cluster_arn, str):
        cluster_name = re.sub(
            r"(?:^arn:aws(?:-[a-z]+)?:elasticache:[\w-]+:[0-9]{12}:cluster:)",
            "",
            cluster_arn,
        )
        return get_cluster_config(resource, cluster_name, session)
    elif isinstance(cluster_arn, list):
        if not re.match(
            r"(?:^arn:aws(?:-[a-z]+)?:elasticache:[\w-]+:[0-9]{12}:cluster:)([\S]+)(?:-[0-9]+)$",
            cluster_arn[0],
        ).groups():
            raise ValueError("Could not match the ARN to a specific Replica Group")
        cluster_name = re.match(
            r"(?:^arn:aws(?:-[a-z]+)?:elasticache:[\w-]+:[0-9]{12}:cluster:)([\S]+)(?:-[0-9]+)$",
            cluster_arn[0],
        ).groups()[0]
        return get_replica_group_config(resource, cluster_name, session)


def lookup_cluster_resource(resource, session):
    """
    Function to find the DB in AWS account

    :param boto3.session.Session session: Boto3 session for clients
    :return:
    """
    elasticache_types = {
        "elasticache:cluster": {
            "regexp": r"(?:^arn:aws(?:-[a-z]+)?:elasticache:[\w-]+:[0-9]{12}:cluster:)([\S]+)$"
        }
    }
    res_type = "elasticache:cluster"
    lookup_session = define_lookup_role_from_info(resource.lookup, session)
    cluster_arn = find_aws_resource_arn_from_tags_api(
        resource.lookup,
        lookup_session,
        res_type,
        types=elasticache_types,
        allow_multi=True,
    )
    if not cluster_arn:
        return None
    cluster_config = return_cluster_config(resource, cluster_arn, lookup_session)
    LOG.debug(cluster_config)
    return cluster_config



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Functions to add to the Cluster template when people want to use SpotFleet for their ECS Cluster.
"""

from troposphere import GetAtt, If, Ref, Select, Sub
from troposphere.applicationautoscaling import (
    ScalableTarget,
    ScalingPolicy,
    StepAdjustment,
    StepScalingPolicyConfiguration,
)
from troposphere.cloudwatch import Alarm
from troposphere.cloudwatch import MetricDimension as CwMetricDimension
from troposphere.ec2 import (
    LaunchTemplate,
    LaunchTemplateConfigs,
    LaunchTemplateOverrides,
    LaunchTemplateSpecification,
    SpotFleet,
    SpotFleetRequestConfigData,
)
from troposphere.iam import Role

from ecs_composex.common import LOG, build_template
from ecs_composex.compute import compute_conditions, compute_params
from ecs_composex.iam import service_role_trust_policy
from ecs_composex.vpc import vpc_params


def add_fleet_role(template):
    """Function to create the IAM Role for the EC2 Spot Fleet

    :param template: source template to add the resources to
    :type template: troposphere.Template

    :returns: role
    :rtype: troposphere.iam.Role
    """
    role = Role(
        "IamFleetRole",
        template=template,
        AssumeRolePolicyDocument=service_role_trust_policy("spotfleet"),
        ManagedPolicyArns=[
            "arn:aws:iam::aws:policy/service-role/AmazonEC2SpotFleetAutoscaleRole",
            "arn:aws:iam::aws:policy/service-role/AmazonEC2SpotFleetTaggingRole",
        ],
    )
    return role


def define_overrides(settings, lt_id, lt_version, spot_config):
    """
    From the list of AZs and the configurations set for spotfleet instances, it will generate an override that
    SpotFleet will use to diversify the compute resources.

    :param ecs_composex.common.settings.ComposeXSettings settings: The settings for execution
    :param lt_id: Launch template ID
    :param lt_version: Launch Template Version
    :param spot_config: SpotFleet configuration for pricing and instance types
    :return: configs
    :type: list
    """
    if isinstance(lt_id, LaunchTemplate):
        template_id = Ref(lt_id)
        template_version = GetAtt(lt_id, "LatestVersionNumber")
    elif not isinstance(lt_id, Ref) or not isinstance(lt_version, Ref):
        raise TypeError(
            "The launch template and/or version are of type",
            lt_id,
            lt_version,
            "Expected",
            LaunchTemplate,
            "Or",
            Ref,
        )
    else:
        template_id = lt_id
        template_version = lt_version

    overrides = []
    configs = []
    LOG.debug(spot_config)
    for count, az in enumerate(settings.aws_azs):
        LOG.debug(az)
        for itype in spot_config["spot_instance_types"]:
            overrides.append(
                LaunchTemplateOverrides(
                    SubnetId=Select(count, Ref(vpc_params.APP_SUBNETS)),
                    WeightedCapacity=spot_config["spot_instance_types"][itype][
                        "weight"
                    ],
                    InstanceType=itype,
                )
            )
    configs.append(
        LaunchTemplateConfigs(
            LaunchTemplateSpecification=LaunchTemplateSpecification(
                LaunchTemplateId=template_id, Version=template_version
            ),
            Overrides=overrides,
        )
    )
    return configs


def add_scaling_policies(template, spot_fleet, role):
    """Function to add Scaling to the SpotFleet

    :param template: source template to add the resources to
    :type template: troposphere.Template
    :param spot_fleet: the SpotFleet for reference
    :type spot_fleet: troposphere.ec2.SpotFleet
    :param role: IAM Role for the SpotFleet
    :type role: troposphere.iam.Role

    :returns: tuple(scale_in_policy, scale_out_policy)
    :rtype: tuple
    """
    target = ScalableTarget(
        f"SpotFleetScalingTarget{spot_fleet.title}",
        template=template,
        MaxCapacity=If(
            compute_conditions.MAX_IS_MIN_T,
            Ref(compute_params.MIN_CAPACITY),
            Ref(compute_params.MAX_CAPACITY),
        ),
        MinCapacity=Ref(compute_params.MIN_CAPACITY),
        ResourceId=Sub(f"spot-fleet-request/${{{spot_fleet.title}}}"),
        RoleARN=GetAtt(role, "Arn"),
        ServiceNamespace="ec2",
        ScalableDimension="ec2:spot-fleet-request:TargetCapacity",
    )

    scale_in_policy = ScalingPolicy(
        f"FleetScalingIn{spot_fleet.title}",
        template=template,
        PolicyName=Sub(f"${{{spot_fleet.title}}}ScalingIn"),
        PolicyType="StepScaling",
        ScalingTargetId=Ref(target),
        StepScalingPolicyConfiguration=StepScalingPolicyConfiguration(
            AdjustmentType="ChangeInCapacity",
            Cooldown=300,
            MetricAggregationType="Average",
            StepAdjustments=[
                StepAdjustment(MetricIntervalLowerBound=10, ScalingAdjustment="-1")
            ],
        ),
    )

    scale_out_policy = ScalingPolicy(
        f"FleetScalingOut{spot_fleet.title}",
        template=template,
        PolicyName=Sub(f"${{{spot_fleet.title}}}CpuAverageScaleOut"),
        PolicyType="StepScaling",
        ScalingTargetId=Ref(target),
        StepScalingPolicyConfiguration=StepScalingPolicyConfiguration(
            AdjustmentType="ChangeInCapacity",
            Cooldown=300,
            MetricAggregationType="Average",
            StepAdjustments=[
                StepAdjustment(MetricIntervalLowerBound=10, ScalingAdjustment="1")
            ],
        ),
    )
    return scale_in_policy, scale_out_policy


def define_default_cw_alarms(template, spot_fleet, scaling_set):
    """
    Function to define CW alarms for the fleet. These default CW Alarms will purely rely on
    EC2 Metrics.

    :param template: Cluster Template
    :type template: troposphere.Template
    :param spot_fleet: The spot fleet the alarms are for
    :type spot_fleet: troposphere.ec2.SpotFleet
    :param scaling_set: Scale In/Out Policies triggered by alarams
    :type scaling_set: tuple

    :returns: void
    """
    Alarm(
        f"LowCpuAverageAlarm{spot_fleet.title}",
        template=template,
        ActionsEnabled=True,
        AlarmActions=[Ref(scaling_set[0])],
        AlarmDescription=Sub(f"LOW CPU USAGE FOR ${{{spot_fleet.title}}}"),
        AlarmName=Sub(f"CPU_AVG_LOW_${{{spot_fleet.title}}}"),
        ComparisonOperator="LessThanOrEqualToThreshold",
        Dimensions=[CwMetricDimension(Name="FleetRequestId", Value=Ref(spot_fleet))],
        EvaluationPeriods=5,
        Period=60,
        Namespace="AWS/EC2Spot",
        MetricName="CPUUtilization",
        Statistic="Average",
        Threshold="25",
        Unit="Percent",
        TreatMissingData="notBreaching",
    )

    Alarm(
        f"HighCpuAverageAlarm{spot_fleet.title}",
        ActionsEnabled=True,
        AlarmActions=[Ref(scaling_set[1])],
        AlarmDescription=Sub(f"LOW CPU USAGE FOR ${{{spot_fleet.title}}}"),
        AlarmName=Sub(f"CPU_AVG_HIGH_${{{spot_fleet.title}}}"),
        ComparisonOperator="GreaterThanOrEqualToThreshold",
        Dimensions=[CwMetricDimension(Name="FleetRequestId", Value=Ref(spot_fleet))],
        EvaluationPeriods=5,
        Period=60,
        Namespace="AWS/EC2Spot",
        MetricName="CPUUtilization",
        Statistic="Average",
        Threshold="65",
        Unit="Percent",
        TreatMissingData="notBreaching",
    )


def define_spot_fleet(template, settings, lt_id, lt_version, spot_config):
    """
    Function to add a spot fleet to the cluster template

    :param ecs_composex.common.settings.ComposeXSettings settings: The settings for execution
    :param template: Source template
    :type template: troposphere.Template

    :returns: NIL
    """
    configs = define_overrides(settings, lt_id, lt_version, spot_config)
    role = add_fleet_role(template)
    fleet = SpotFleet(
        "EcsClusterFleet",
        template=template,
        SpotFleetRequestConfigData=SpotFleetRequestConfigData(
            AllocationStrategy="diversified",
            ExcessCapacityTerminationPolicy="default",
            IamFleetRole=GetAtt(role, "Arn"),
            InstanceInterruptionBehavior="terminate",
            TargetCapacity=Ref(compute_params.TARGET_CAPACITY),
            Type="maintain",
            SpotPrice=str(spot_config["bid_price"]),
            ReplaceUnhealthyInstances=True,
            LaunchTemplateConfigs=configs,
        ),
    )
    scaling_set = add_scaling_policies(template, fleet, role)
    define_default_cw_alarms(template, fleet, scaling_set)


def generate_spot_fleet_template(settings, spot_config):
    """
    Generates a standalone template for SpotFleet
    """
    template = build_template(
        "Template For SpotFleet As Part Of EcsCluster",
        [
            compute_params.LAUNCH_TEMPLATE_ID,
            compute_params.LAUNCH_TEMPLATE_VersionNumber,
            compute_params.MIN_CAPACITY,
            compute_params.MAX_CAPACITY,
            compute_params.TARGET_CAPACITY,
            vpc_params.APP_SUBNETS,
        ],
    )
    template.add_condition(
        compute_conditions.MAX_IS_MIN_T, compute_conditions.MAX_IS_MIN
    )
    lt_id = Ref(compute_params.LAUNCH_TEMPLATE_ID)
    lt_version = Ref(compute_params.LAUNCH_TEMPLATE_VersionNumber)
    define_spot_fleet(template, settings, lt_id, lt_version, spot_config)
    return template



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from json import loads
from os import path


def get_access_types():
    with open(
        f"{path.abspath(path.dirname(__file__))}/s3_perms.json",
        "r",
        encoding="utf-8-sig",
    ) as perms_fd:
        return loads(perms_fd.read())


ACCESS_TYPES = get_access_types()



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Parameters related to the VPC settings. Used by ecs_composex.vpc and others
"""

from ecs_composex.common.cfn_params import Parameter

VPC_TYPE = "AWS::EC2::VPC::Id"
AMI_TYPE = "AWS::EC2::Image::Id"
SUBNET_TYPE = "AWS::EC2::Subnet::Id"
SUBNETS_TYPE = "List<AWS::EC2::Subnet::Id>"
R53_TYPE = "AWS::Route53::HostedZone::Id"
AZS_TYPE = "List<AWS::EC2::AvailabilityZone::Name>"
SG_ID_TYPE = "AWS::EC2::SecurityGroup::Id"
SG_NAME_TYPE = "AWS::EC2::SecurityGroup::GroupName"

DEFAULT_VPC_CIDR = "100.64.72.0/24"
DEFAULT_SINGLE_NAT = True


VPC_T = "Vpc"
IGW_T = "InternetGatewayV4"

RES_KEY = "vpc"
VPC_ID_T = "VpcId"
VPC_ID = Parameter(VPC_ID_T, Type=VPC_TYPE)

VPC_CIDR_T = "VpcCidr"
VPC_CIDR = Parameter(VPC_CIDR_T, Type="String", Default=DEFAULT_VPC_CIDR)

VPC_SINGLE_NAT_T = "SingleNat"
VPC_SINGLE_NAT = Parameter(VPC_SINGLE_NAT_T, Type="String", Default="True")

STORAGE_SUBNETS_CIDR_T = "StorageSubnetsCidr"
STORAGE_SUBNETS_CIDR = Parameter(STORAGE_SUBNETS_CIDR_T, Type="CommaDelimitedList")
STORAGE_SUBNETS_T = "StorageSubnets"
STORAGE_SUBNETS = Parameter(STORAGE_SUBNETS_T, Type=SUBNETS_TYPE)

BACKEND_SUBNETS_CIDR_T = "BackendSubnetsCidr"
BACKEND_SUBNETS_CIDR = Parameter(BACKEND_SUBNETS_CIDR_T, Type="CommaDelimitedList")
BACKEND_SUBNETS_T = "BackendSubnets"
BACKEND_SUBNETS = Parameter(BACKEND_SUBNETS_T, Type=SUBNETS_TYPE)

PUBLIC_SUBNETS_CIDR_T = "PublicSubnetsCidr"
PUBLIC_SUBNETS_CIDR = Parameter(PUBLIC_SUBNETS_CIDR_T, Type="CommaDelimitedList")
PUBLIC_SUBNETS_T = "PublicSubnets"
PUBLIC_SUBNETS = Parameter(PUBLIC_SUBNETS_T, Type=SUBNETS_TYPE)

APP_SUBNETS_CIDR_T = "AppSubnetsCidr"
APP_SUBNETS_CIDR = Parameter(APP_SUBNETS_CIDR_T, Type="CommaDelimitedList")
APP_SUBNETS_T = "AppSubnets"
APP_SUBNETS = Parameter(APP_SUBNETS_T, Type=SUBNETS_TYPE)

USE_SUB_ZONE_T = "UseSubDnsZone"
USE_SUB_ZONE = Parameter(
    USE_SUB_ZONE_T,
    Type="String",
    AllowedValues=["True", "False"],
    Default="True",
)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Package to handle recurring Secrets tasks
"""

from troposphere import Parameter, Ref, Sub
from troposphere.docdb import DBCluster as DocdbCluster
from troposphere.docdb import DBInstance as DocdbInstance
from troposphere.rds import DBCluster as RdsCluster
from troposphere.rds import DBInstance as RdsInstance
from troposphere.secretsmanager import (
    GenerateSecretString,
    Secret,
    SecretTargetAttachment,
)

from ecs_composex.common import add_parameters


def add_db_secret(template, resource_title):
    """
    Function to add a Secrets Manager secret that will be associated with the DB

    :param template.Template template: The template to add the secret to.
    :param str resource_title: The Logical name of the resource associated to that secret
    """
    username = Parameter(
        f"{resource_title}Username",
        Type="String",
        MinLength=3,
        MaxLength=16,
        Default="dbadmin",
    )
    password_length = Parameter(
        f"{resource_title}PasswordLength",
        Type="Number",
        MinValue=8,
        MaxValue=32,
        Default=16,
    )
    add_parameters(template, [username, password_length])
    secret = Secret(
        f"{resource_title}Secret",
        template=template,
        GenerateSecretString=GenerateSecretString(
            SecretStringTemplate=Sub(f'{{"username":"${{{username.title}}}"}}'),
            GenerateStringKey="password",
            ExcludeCharacters="<>%`|;,.",
            ExcludePunctuation=True,
            ExcludeLowercase=False,
            ExcludeUppercase=False,
            IncludeSpace=False,
            RequireEachIncludedType=True,
            PasswordLength=Ref(password_length),
        ),
    )
    return secret


def add_db_dependency(resource, secret):
    if hasattr(resource, "DependsOn") and secret.title not in resource.DependsOn:
        resource.DependsOn.append(secret.title)
    elif not hasattr(resource, "DependsOn"):
        setattr(resource, "DependsOn", [secret.title])


def attach_to_secret_to_resource(template, resource, secret):
    """
    Function to associate a secret to a resource
    :param troposphere.Template template:
    :param resource: The resource we can link the secret to.
    :param secret: The secret to attach to the resource
    :return:
    """
    if not isinstance(resource, (RdsCluster, RdsInstance, DocdbCluster, DocdbInstance)):
        raise TypeError(
            "The resource to attach can only be one of ",
            (RdsCluster, RdsInstance, DocdbCluster, DocdbInstance),
            "Got",
            type(resource),
        )
    SecretTargetAttachment(
        f"{resource.title}SecretAttachment",
        template=template,
        DependsOn=[resource.title, secret.title],
        TargetType=resource.resource_type,
        SecretId=Ref(secret),
        TargetId=Ref(resource),
    )



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Package for x-dns
"""

from compose_x_common.compose_x_common import keyisset
from troposphere import AWS_NO_VALUE, FindInMap, GetAtt, Ref, Sub, Tags
from troposphere.route53 import HostedZone, HostedZoneConfiguration
from troposphere.servicediscovery import PrivateDnsNamespace as VpcSpace

from ecs_composex.common import LOG, add_parameters, cfn_params
from ecs_composex.common.aws import get_cross_role_session
from ecs_composex.common.stacks import ComposeXStack
from ecs_composex.dns import dns_params
from ecs_composex.dns.dns_lookup import lookup_namespace


class DnsZone(object):
    """
    Class to represent a zone
    """

    key = None
    nested_name_parameter = None
    nested_id_parameter = None

    def __init__(self, definition):
        """
        Init class
        """
        if not keyisset("Name", definition):
            raise KeyError(
                "You must provide the DNS Name you wish to use for this setting."
            )
        self.root_params = {}
        self.nested_params = {}
        self.dns_mapping = {}
        self.cfn_resource = None
        self.create_zone = False
        self.name_value = Ref(AWS_NO_VALUE)
        self.id_value = Ref(AWS_NO_VALUE)
        self.definition = definition
        self.lookup = (
            self.definition["Lookup"] if keyisset("Lookup", self.definition) else None
        )
        self.use = self.definition["Use"] if keyisset("Use", self.definition) else None
        if self.lookup or self.use:
            self.name_value = FindInMap(
                "Dns", self.key, self.nested_name_parameter.title
            )
            self.id_value = FindInMap("Dns", self.key, self.nested_id_parameter.title)
            self.nested_params.update(
                {
                    self.nested_id_parameter.title: self.id_value,
                    self.nested_name_parameter.title: self.name_value,
                }
            )
        self.name = self.definition["Name"]

    def setup(self, settings):
        if self.lookup:
            self.set_zone_from_lookup(settings)
        elif self.use:
            self.set_zone_from_use()
        else:
            self.create_zone = True
            self.set_zone_default()

    def update_nested_stack_parameters(self, nested_stack):
        """
        Method to updated a nested stack parameters to point to the right value for DNS Settings.

        :param ecs_composex.common.stacks.ComposeXStack nested_stack:
        :return:
        """
        if self.nested_id_parameter.title in nested_stack.stack_template.parameters:
            add_parameters(nested_stack.stack_template, [self.nested_id_parameter])
            nested_stack.Parameters.update(
                {self.nested_id_parameter.title: self.id_value}
            )

        if self.nested_name_parameter.title in nested_stack.stack_template.parameters:
            add_parameters(nested_stack.stack_template, [self.nested_name_parameter])
            nested_stack.Parameters.update(
                {self.nested_name_parameter.title: self.name_value}
            )

    def set_zone_from_lookup(self, settings):
        if keyisset("RoleArn", self.lookup):
            session = get_cross_role_session(
                settings.session, arn=self.lookup["RoleArn"]
            )
        else:
            session = settings.session
        namespace_info = lookup_namespace(self, session)
        if not namespace_info["ZoneTld"].find(self.name) == 0:
            raise ValueError(
                "Zone name provided does not match the value looked up. Got",
                self.name,
                "Resolved via ID",
                namespace_info["ZoneTld"],
            )
        self.dns_mapping.update(
            {
                self.key: {
                    self.nested_name_parameter.title: namespace_info["ZoneTld"],
                    self.nested_id_parameter.title: namespace_info["ZoneId"],
                }
            },
        )
        if hasattr(self, "zone_id_parameter") and keyisset("Route53ID", namespace_info):
            self.dns_mapping[self.key].update(
                {self.zone_id_parameter.title: namespace_info["Route53ID"]}
            )
        self.id_value = FindInMap("Dns", self.key, self.nested_id_parameter.title)
        self.name_value = FindInMap("Dns", self.key, self.nested_name_parameter.title)

    def set_zone_from_use(self):
        """
        Method to create the CFN Mapping for Use on DNS Zone.
        :return:
        """
        self.dns_mapping.update(
            {
                self.key: {
                    self.nested_id_parameter.title: self.use,
                    self.nested_name_parameter.title: self.name,
                }
            }
        )
        self.id_value = FindInMap("Dns", self.key, self.nested_id_parameter.title)
        self.name_value = FindInMap("Dns", self.key, self.nested_name_parameter.title)

    def set_zone_default(self):
        self.dns_mapping.update(
            {self.key: {self.nested_name_parameter.title: self.name}}
        )
        self.nested_params.update(
            {
                self.nested_name_parameter.title: FindInMap(
                    "Dns", self.key, self.nested_name_parameter.title
                ),
            }
        )


class PrivateNamespace(DnsZone):
    """
    Class to handle the private DNS Namespace associated with the VPC used for Service discovery
    """

    key = "PrivateNamespace"
    nested_name_parameter = dns_params.PRIVATE_DNS_ZONE_NAME
    nested_id_parameter = dns_params.PRIVATE_NAMESPACE_ID
    zone_id_parameter = dns_params.PRIVATE_DNS_ZONE_ID

    def __init__(self, definition):
        """
        Init class
        """
        super().__init__(definition)

    def add_zone(self, template, vpc):
        self.cfn_resource = VpcSpace(
            cfn_params.PRIVATE_MAP_TITLE,
            template=template,
            Description=Sub(r"CloudMap VpcNamespace for ${AWS::StackName}"),
            Name=self.name,
            Vpc=vpc,
            DependsOn=[]
            if isinstance(vpc, (FindInMap, Sub, Ref))
            else [cfn_params.VPC_STACK_NAME],
        )
        self.id_value = GetAtt(self.cfn_resource, "Id")
        self.name_value = self.name


class PublicZone(DnsZone):
    """
    Class to represent the Public DNS Zone used for inbound. This is a Route53 DNS Zone.
    """

    key = "PublicZone"
    nested_name_parameter = dns_params.PUBLIC_DNS_ZONE_NAME
    nested_id_parameter = dns_params.PUBLIC_DNS_ZONE_ID

    def __init__(self, definition):
        """ """
        super().__init__(definition)

    def add_zone(self, template):
        self.cfn_resource = HostedZone(
            cfn_params.PUBLIC_ZONE_TITLE,
            Name=self.name,
            template=template,
            HostedZoneTags=Tags(CreatedByComposeX="True", PublicZone="True"),
            HostedZoneConfig=HostedZoneConfiguration(
                Comment=Sub("Public DNS Zone for ${AWS::StackName}")
            ),
        )
        self.id_value = Ref(self.cfn_resource)
        self.name_value = self.name


class DnsSettings(object):
    """
    Class to ingest the x-dns settings
    """

    main_key = "x-dns"
    private_namespace_key = "PrivateNamespace"
    public_namespace_key = "PublicNamespace"
    public_zone_key = "PublicZone"
    supported_keys = [private_namespace_key, public_namespace_key]
    default_private_name = dns_params.PRIVATE_DNS_ZONE_NAME.Default

    def __init__(self, root_stack, settings, vpc):
        """
        Method to initialize DnsSettings class
        :param ecs_composex.common.settings.ComposeXSettings settings: Settings for execution
        """
        self.private_zone_name = dns_params.PRIVATE_DNS_ZONE_NAME.Default
        self.public_zone_name = dns_params.PUBLIC_DNS_ZONE_NAME.Default
        self.public_zone = None
        self.private_zone = None
        self.dns_mapping = {
            PrivateNamespace.key: {
                PrivateNamespace.nested_name_parameter.title: self.default_private_name
            }
        }
        dns_settings = (
            {}
            if not keyisset(self.main_key, settings.compose_content)
            else settings.compose_content[self.main_key]
        )

        if not dns_settings and not settings.requires_private_namespace:
            LOG.warning(
                f"No {self.main_key} defined and no mesh nor service registry defined. Skipping"
            )
            return
        elif not dns_settings and settings.requires_private_namespace:
            LOG.info(
                f"No {self.main_key} defined but a private namespace is required. Using default"
            )
            dns_settings = {PrivateNamespace.key: {"Name": self.default_private_name}}

        if keyisset(PrivateNamespace.key, dns_settings):
            self.private_zone = PrivateNamespace(dns_settings[PrivateNamespace.key])
            self.private_zone.setup(settings)
            if self.private_zone.create_zone:
                self.private_zone.add_zone(root_stack.stack_template, vpc)
            self.dns_mapping.update(self.private_zone.dns_mapping)

        if keyisset(PublicZone.key, dns_settings):
            self.public_zone = PublicZone(dns_settings[PublicZone.key])
            self.public_zone.setup(settings)
            if self.public_zone.create_zone:
                self.public_zone.add_zone(root_stack.stack_template)
            self.dns_mapping.update(self.public_zone.dns_mapping)

        if dns_settings:
            root_stack.stack_template.add_mapping("Dns", self.dns_mapping)

    def associate_settings_to_nested_stacks(self, root_stack):
        """
        Method to apply the public and private zone parameters to nested stacks if they needed it.

        :param ecs_composex.common.stacks.ComposeXStack root_stack:
        :return:
        """
        for (
            nested_stack_name,
            nested_stack,
        ) in root_stack.stack_template.resources.items():
            if isinstance(nested_stack, ComposeXStack) or issubclass(
                type(nested_stack), ComposeXStack
            ):
                if self.private_zone:
                    self.private_zone.update_nested_stack_parameters(nested_stack)
                if self.public_zone:
                    self.public_zone.update_nested_stack_parameters(nested_stack)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Cluster parameters for CFN
This is a crucial part as all the titles, maked `_T` are string which are then used the same way
across all imports, which gives consistency for CFN to use the same names,
which it heavily relies onto.

You can change the names *values* so you like so long as you keep it [a-zA-Z0-9]
"""

from ecs_composex.common.cfn_params import Parameter
from ecs_composex.ecs.ecs_params import CLUSTER_NAME_T

HOST_ROLE_T = "EcsHostsRole"
HOST_PROFILE_T = "EcsHostsInstanceProfile"
NODES_SG_T = "EcsHostsSg"

CLUSTER_NAME = Parameter(CLUSTER_NAME_T, Type="String")

ECS_AMI_ID_T = "EcsAmiId"
ECS_AMI_ID = Parameter(
    ECS_AMI_ID_T,
    Type="AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>",
    Default="/aws/service/ecs/optimized-ami/amazon-linux-2/recommended/image_id",
)

MAX_CAPACITY_T = "EcsMaxCapacity"
MAX_CAPACITY = Parameter(MAX_CAPACITY_T, Type="Number", Default=1)

MIN_CAPACITY_T = "EcsMinCapacity"
MIN_CAPACITY = Parameter(MIN_CAPACITY_T, Type="Number", MinValue=1, Default=1)

TARGET_CAPACITY_T = "EcsTargetCapacity"
TARGET_CAPACITY = Parameter(
    TARGET_CAPACITY_T, Type="Number", Default=MIN_CAPACITY.Default
)

LAUNCH_TEMPLATE_ID_T = "LaunchTemplateId"
LAUNCH_TEMPLATE_ID = Parameter(
    LAUNCH_TEMPLATE_ID_T, Type="String", Description="ID Of the Launch Template"
)

LAUNCH_TEMPLATE_VersionNumber_T = "LaunchTemplateVersionNumber"
LAUNCH_TEMPLATE_VersionNumber = Parameter(
    LAUNCH_TEMPLATE_VersionNumber_T,
    Type="Number",
    Description="VersionNumber Of the Launch Template",
)



#   -*- coding: utf-8 -*-
#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2021 John Mille <john@compose-x.io>

from os import path

from ecs_composex.common.compose_resources import Parameter
from ecs_composex.common.ecs_composex import X_KEY

MOD_KEY = path.basename(path.dirname(path.abspath(__file__)))
RES_KEY = f"{X_KEY}{MOD_KEY}"

PROFILER_NAME = Parameter("ProfileName", Type="String")
PROFILER_ARN = Parameter("ProfileArn", return_value="Arn", Type="String")



#   -*- coding: utf-8 -*-
#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to add Prometheus scraper for ECS tasks.
"""
import json
import re

import yaml

try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper

from troposphere import AWS_ACCOUNT_ID, AWS_PARTITION, AWS_REGION, AWS_STACK_NAME, Sub
from troposphere.ecs import Secret
from troposphere.iam import Policy
from troposphere.ssm import Parameter as SSMParameter

from ecs_composex.common import keyisset
from ecs_composex.common.cfn_params import Parameter
from ecs_composex.common.compose_services import ComposeService
from ecs_composex.ecs import ecs_params

CW_IMAGE_PARAMETER = Parameter(
    "CloudwatchAgentImage",
    Type="String",
    Default="public.ecr.aws/cloudwatch-agent/cloudwatch-agent:latest",
)

NGINX_EXPORTER_IMAGE_PARAMETER = Parameter(
    "NginxPrometheusExporterImage",
    Type="String",
    Default="public.ecr.aws/compose-x/nginx-prometheus-exporter:0.9.0",
)

METRICS_DEFAULT_PATH = r"/metrics"


def set_cw_prometheus_config_parameter(family):
    """
    Function to add the SSM Parameter representing the Prometheus scrapper config
    :param ecs_composex.common.compose_services.ComposeFamily family:
    :return: parameter
    :rtype: troposphere.ssm.Parameter
    """
    value_py = {
        "global": {
            "scrape_interval": "1m",
            "scrape_timeout": "10s",
        },
        "scrape_configs": [
            {
                "job_name": "cwagent-ecs-file-sd-config",
                "sample_limit": 10000,
                "file_sd_configs": [{"files": ["/tmp/cwagent_ecs_auto_sd.yaml"]}],
            }
        ],
    }

    parameter = SSMParameter(
        f"{family.logical_name}SSMPrometheusConfig",
        Tier="Standard",
        Type="String",
        Name=Sub(
            f"/ecs/config/prometheus/${{{ecs_params.CLUSTER_NAME.title}}}/${{{ecs_params.SERVICE_NAME_T}}}"
        ),
        Description=Sub(
            f"Prometheus Scraping SSM Parameter for ECS Cluster: ${{{ecs_params.CLUSTER_NAME.title}}}"
        ),
        Value=yaml.dump(value_py, Dumper=Dumper),
    )
    family.template.add_resource(parameter)
    return parameter


def generate_ecs_sd_service_name_pattern(family_name):
    """
    Generate the ecs_service_discovery configuration for a given set of set of ECS Task Families

    :param str family_name:
    :return:
    """
    task_def_re = re.compile(
        r"(.*:task-definition/)|(arn:aws(?:[\S]+)?:ecs:[\S]+:\d{12}:task-definition/)"
    )
    if not task_def_re.match(family_name):
        task_name = f".*:task-definition/.*{family_name}"
    else:
        task_name = family_name
    return task_name


def get_ecs_envoy_processor(envoy_container_name=None):
    """
    Function to return the envoy EMF configuration

    :param str envoy_container_name:
    :return:
    """
    if envoy_container_name is None:
        envoy_container_name = r"envoy"
    return [
        {
            "source_labels": ["container_name"],
            "label_matcher": f"^{envoy_container_name}$",
            "dimensions": [["ClusterName", "TaskDefinitionFamily"]],
            "metric_selectors": [
                "^envoy_http_downstream_rq_(total|xx)$",
                "^envoy_cluster_upstream_cx_(r|t)x_bytes_total$",
                "^envoy_cluster_membership_(healthy|total)$",
                "^envoy_server_memory_(allocated|heap_size)$",
                "^envoy_cluster_upstream_cx_(connect_timeout|destroy_local_with_active_rq)$",
                "^envoy_cluster_upstream_rq_(pending_failure_eject|"
                "pending_overflow|timeout|per_try_timeout|rx_reset|maintenance_mode)$",
                "^envoy_http_downstream_cx_destroy_remote_active_rq$",
                "^envoy_cluster_upstream_flow_control_(paused_reading_total|"
                "resumed_reading_total|backed_up_total|drained_total)$",
                "^envoy_cluster_upstream_rq_retry$",
                "^envoy_cluster_upstream_rq_retry_(success|overflow)$",
                "^envoy_server_(version|uptime|live)$",
            ],
        },
        {
            "source_labels": ["container_name"],
            "label_matcher": f"^{envoy_container_name}$",
            "dimensions": [
                [
                    "ClusterName",
                    "TaskDefinitionFamily",
                    "envoy_http_conn_manager_prefix",
                    "envoy_response_code_class",
                ]
            ],
            "metric_selectors": ["^envoy_http_downstream_rq_xx$"],
        },
    ]


def define_nginx_exporter_sidecar(family):
    """
    Function to define the NGINX Exporter sidecar

    :param ecs_composex.common.compose_services.ComposeFamily family:

    :return:
    """
    nginx_prom_exporter_agent_service_config = {
        "image": NGINX_EXPORTER_IMAGE_PARAMETER.Default,
        "deploy": {
            "resources": {"limits": {"cpus": 0.1, "memory": "64M"}},
            "labels": {"ecs.task.family": family.name},
        },
        "labels": {
            "container_name": "nginx-prometheus-exporter",
            "ECS_PROMETHEUS_EXPORTER_PORT": 9113,
            "job": "nginx-prometheus-exporter",
        },
        "ports": [{"target": 9113, "protocol": "tcp"}],
        "depends_on": [
            service.name for service in family.services if not service.is_aws_sidecar
        ],
    }
    nginx_prom_exporter_service = ComposeService(
        "nginx-prometheus-exporter", nginx_prom_exporter_agent_service_config
    )
    nginx_prom_exporter_service.is_aws_sidecar = True

    family.add_service(nginx_prom_exporter_service)
    family.refresh()


def get_ngnix_processor(
    family,
    ecs_sd_config,
    nginx_config,
):
    labels = (
        ["job"]
        if not keyisset("source_labels", nginx_config)
        else nginx_config["source_labels"]
    )
    default_matcher = (
        r"^.*nginx.*$"
        if not keyisset("label_matcher", nginx_config)
        else nginx_config["label_matcher"]
    )
    nginx_metrics = [
        {
            "source_labels": labels,
            "label_matcher": default_matcher,
            "dimensions": [["ClusterName", "TaskDefinitionFamily", "ServiceName"]],
            "metric_selectors": ["^nginx_.*$"],
        }
    ]
    ecs_sd_config["task_definition_list"].append(
        {
            "sd_job_name": "nginx-prometheus-exporter",
            "sd_metrics_path": METRICS_DEFAULT_PATH
            if not keyisset("ExporterPath", nginx_config)
            else nginx_config["ExporterPath"],
            "sd_metrics_ports": "9113"
            if not keyisset("ExporterPort", nginx_config)
            else str(nginx_config["ExporterPort"]),
            "sd_task_definition_arn_pattern": generate_ecs_sd_service_name_pattern(
                family.name
            ),
        },
    )
    ecs_sd_config["service_name_list_for_tasks"].append(
        {
            "sd_job_name": "nginx-prometheus-exporter",
            "sd_metrics_path": METRICS_DEFAULT_PATH
            if not keyisset("ExporterPath", nginx_config)
            else nginx_config["ExporterPath"],
            "sd_metrics_ports": "9113"
            if not keyisset("ExporterPort", nginx_config)
            else str(nginx_config["ExporterPort"]),
            "sd_service_name_pattern": f"^.*${{{AWS_STACK_NAME}}}.*$",
        }
    )
    return nginx_metrics


def get_jmx_processor(family, ecs_sd_config, jmx_config):
    labels = (
        ["job"]
        if not keyisset("source_labels", jmx_config)
        else jmx_config["source_labels"]
    )
    default_matcher = (
        r"^.*jmx.*$"
        if not keyisset("label_matcher", jmx_config)
        else jmx_config["label_matcher"]
    )
    jmx_metrics = [
        {
            "source_labels": labels,
            "label_matcher": default_matcher,
            "dimensions": [["ClusterName", "TaskDefinitionFamily"]],
            "metric_selectors": [
                "^jvm_threads_(current|daemon)$",
                "^jvm_classes_loaded$",
                "^java_lang_operatingsystem_(freephysicalmemorysize|totalphysicalmemorysize|freeswapspacesize"
                "|totalswapspacesize|systemcpuload|processcpuload|availableprocessors|openfiledescriptorcount)$",
                "^catalina_manager_(rejectedsessions|activesessions)$",
                "^jvm_gc_collection_seconds_(count|sum)$",
                "^catalina_globalrequestprocessor_(bytesreceived|bytessent|requestcount|errorcount|processingtime)$",
            ],
        },
        {
            "source_labels": labels,
            "label_matcher": default_matcher,
            "dimensions": [["ClusterName", "TaskDefinitionFamily", "area"]],
            "metric_selectors": ["^jvm_memory_bytes_used$"],
        },
        {
            "source_labels": labels,
            "label_matcher": default_matcher,
            "dimensions": [["ClusterName", "TaskDefinitionFamily", "pool"]],
            "metric_selectors": ["^jvm_memory_pool_bytes_used$"],
        },
    ]
    ecs_sd_config["task_definition_list"].append(
        {
            "sd_job_name": "javajmx-prometheus-exporter",
            "sd_metrics_path": METRICS_DEFAULT_PATH
            if not keyisset("ExporterPath", jmx_config)
            else jmx_config["ExporterPath"],
            "sd_metrics_ports": "9404"
            if not keyisset("ExporterPort", jmx_config)
            else str(jmx_config["ExporterPort"]),
            "sd_task_definition_arn_pattern": generate_ecs_sd_service_name_pattern(
                family.name
            ),
        },
    )
    return jmx_metrics


def process_custom_rules(family, ecs_sd_config, options, emf_processors):
    """
    Func
    :param ecs_composex.common.compose_services.ComposeFamily family:
    :param dict ecs_sd_config:
    :param dict options:
    :param dict emf_processors:
    :return:
    """
    custom_rules = options["CustomRules"]
    for count, rule in enumerate(custom_rules):
        emf_processors["metric_declaration"] += rule["EmfProcessors"]
        ecs_sd_config["service_name_list_for_tasks"].append(
            {
                "sd_job_name": f"service-def-{family.logical_name}-custom-sd-{count}",
                "sd_metrics_path": METRICS_DEFAULT_PATH
                if not keyisset("ExporterPath", rule)
                else rule["ExporterPath"],
                "sd_metrics_ports": str(rule["ExporterPort"]),
                "sd_service_name_pattern": f"^.*${{{AWS_STACK_NAME}}}.*$",
            }
        )
        ecs_sd_config["task_definition_list"].append(
            {
                "sd_job_name": f"task-def-{family.logical_name}-custom-sd-{count}",
                "sd_metrics_path": METRICS_DEFAULT_PATH
                if not keyisset("ExporterPath", rule)
                else rule["ExporterPath"],
                "sd_metrics_ports": str(rule["ExporterPort"]),
                "sd_task_definition_arn_pattern": generate_ecs_sd_service_name_pattern(
                    family.name
                ),
            },
        )


def generate_emf_processors(family, ecs_sd_config, **options):
    metrics_key = "metric_declaration"
    emf_processors = {
        "metric_declaration_dedup": True,
        metrics_key: [],
    }
    if keyisset("CollectForAppMesh", options):
        emf_processors[metrics_key] += get_ecs_envoy_processor()
    if keyisset("CollectForJavaJmx", options):
        emf_processors[metrics_key] += get_jmx_processor(
            family, ecs_sd_config, options["CollectForJavaJmx"]
        )
    if keyisset("CollectForNginx", options):
        emf_processors[metrics_key] += get_ngnix_processor(
            family, ecs_sd_config, options["CollectForNginx"]
        )
        if keyisset("AutoAddNginxPrometheusExporter", options):
            define_nginx_exporter_sidecar(family)
    if keyisset("CustomRules", options):
        process_custom_rules(family, ecs_sd_config, options, emf_processors)
    return emf_processors


def set_cw_config_parameter(family, **options):
    """
    Function to add the SSM Parameter representing the Prometheus scrapper config

    :param ecs_composex.common.compose_services.ComposeFamily family:
    :return: parameter
    :rtype: troposphere.ssm.Parameter
    """

    value_py = {
        "logs": {
            "metrics_collected": {
                "prometheus": {
                    "prometheus_config_path": "env:PROMETHEUS_CONFIG_CONTENT",
                    "emf_processor": {
                        "metric_declaration": [],
                    },
                }
            },
            "force_flush_interval": 5,
        }
    }
    if keyisset("EnableCWAgentDebug", options):
        value_py["agent"] = {"debug": True}
    ecs_sd_config = {
        "sd_frequency": "1m",
        "sd_result_file": "/tmp/cwagent_ecs_auto_sd.yaml",
        "docker_label": {},
        "task_definition_list": [],
        "service_name_list_for_tasks": [],
    }
    value_py["logs"]["metrics_collected"]["prometheus"][
        "ecs_service_discovery"
    ] = ecs_sd_config
    if (
        keyisset("CollectForAppMesh", options)
        or keyisset("CollectForJavaJmx", options)
        or keyisset("CollectForNginx", options)
    ):
        emf_processors = generate_emf_processors(family, ecs_sd_config, **options)
        value_py["logs"]["metrics_collected"]["prometheus"][
            "emf_processor"
        ] = emf_processors
    parameter = SSMParameter(
        f"{family.logical_name}SSMCWAgentPrometheusConfig",
        Tier="Intelligent-Tiering",
        Type="String",
        Name=Sub(
            f"/ecs/config/cw_agent_config/${{{ecs_params.CLUSTER_NAME.title}}}/${{{ecs_params.SERVICE_NAME_T}}}"
        ),
        Description=Sub(
            f"Prometheus Scraping SSM Parameter for ECS Cluster: ${{{ecs_params.CLUSTER_NAME.title}}}"
        ),
        Value=Sub(json.dumps(value_py, ensure_ascii=True, sort_keys=True, indent=2)),
    )
    family.template.add_resource(parameter)
    return parameter


def define_cloudwatch_agent(family, cw_prometheus_config, cw_agent_config):
    """
    Function to define the CW Agent image task definition

    :param ecs_composex.common.compose_services.ComposeFamily family:
    :param cw_prometheus_config:
    :param cw_agent_config:
    :return:
    """
    cw_agent_service_config = {
        "image": CW_IMAGE_PARAMETER.Default,
        "deploy": {
            "resources": {"limits": {"cpus": 0.1, "memory": "256M"}},
            "labels": {"ecs.task.family": family.name},
        },
        "labels": {"container_name": "cw-agent"},
        "depends_on": [
            service.name for service in family.services if not service.is_aws_sidecar
        ],
    }
    cw_service = ComposeService("cw_agent", cw_agent_service_config)
    cw_service.is_aws_sidecar = True
    secrets = [
        Secret(
            Name="PROMETHEUS_CONFIG_CONTENT",
            ValueFrom=Sub(
                f"arn:${{{AWS_PARTITION}}}:ssm:${{{AWS_REGION}}}:${{{AWS_ACCOUNT_ID}}}"
                f":parameter${{{cw_prometheus_config.title}}}"
            ),
        ),
        Secret(
            Name="CW_CONFIG_CONTENT",
            ValueFrom=Sub(
                f"arn:${{{AWS_PARTITION}}}:ssm:${{{AWS_REGION}}}:${{{AWS_ACCOUNT_ID}}}"
                f":parameter${{{cw_agent_config.title}}}"
            ),
        ),
    ]
    if hasattr(cw_service.container_definition, "Secrets"):
        s_secrets = getattr(cw_service.container_definition, "Secrets")
        s_secrets += secrets
    else:
        setattr(cw_service.container_definition, "Secrets", secrets)
    return cw_service


def set_ecs_cw_policy(prometheus_parameter, cw_config_parameter):
    """
    Renders the IAM policy to grant the TaskRole access to CW, ECS and SSM Parameters

    :param troposphere.ssm.Parameter prometheus_parameter:
    :param troposphere.ssm.Parameter cw_config_parameter:
    :return: The IAM policy
    :rtype: troposphere.iam.Policy
    """
    ecs_sd_policy = Policy(
        PolicyName="CWAgentAccessForPrometheusScraping",
        PolicyDocument={
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "EnableCreationAndManagementOfPrometheusLogEvents",
                    "Effect": "Allow",
                    "Action": ["logs:GetLogEvents", "logs:PutLogEvents"],
                    "Resource": Sub(
                        f"arn:${{{AWS_PARTITION}}}:logs:*:${{{AWS_ACCOUNT_ID}}}:"
                        "log-group:/aws/ecs/containerinsights/*:log-stream:*"
                    ),
                },
                {
                    "Sid": "EnableCreationAndManagementOfPrometheusCloudwatchLogGroupsAndStreams",
                    "Effect": "Allow",
                    "Action": [
                        "logs:CreateLogStream",
                        "logs:DescribeLogStreams",
                        "logs:PutRetentionPolicy",
                        "logs:CreateLogGroup",
                    ],
                    "Resource": Sub(
                        f"arn:${{{AWS_PARTITION}}}:logs:*:${{{AWS_ACCOUNT_ID}}}:"
                        "log-group:/aws/ecs/containerinsights/*"
                    ),
                },
                {
                    "Sid": "ECSTaskDefinitionsAccess",
                    "Effect": "Allow",
                    "Action": ["ecs:DescribeTaskDefinition"],
                    "Resource": "*",
                },
                {
                    "Sid": "ServiceDiscoveryAccess",
                    "Effect": "Allow",
                    "Action": [
                        "ecs:DescribeTasks",
                        "ecs:ListTasks",
                        "ecs:DescribeContainerInstances",
                        "ecs:DescribeServices",
                        "ecs:ListServices",
                    ],
                    "Resource": "*",
                    "Condition": {
                        "ArnEquals": {
                            "ecs:cluster": Sub(
                                f"arn:${{{AWS_PARTITION}}}:ecs:${{{AWS_REGION}}}:${{{AWS_ACCOUNT_ID}}}"
                                f":cluster/${{{ecs_params.CLUSTER_NAME.title}}}"
                            )
                        }
                    },
                },
                {
                    "Sid": "ExtractFromCloudWatchAgentServerPolicy",
                    "Effect": "Allow",
                    "Action": ["ssm:GetParameter*"],
                    "Resource": [
                        Sub(
                            "arn:aws:ssm:*:${AWS::AccountId}:parameter/AmazonCloudWatch-*"
                        ),
                        Sub(
                            f"arn:${{{AWS_PARTITION}}}:ssm:${{{AWS_REGION}}}:${{{AWS_ACCOUNT_ID}}}"
                            f":parameter${{{prometheus_parameter.title}}}"
                        ),
                        Sub(
                            f"arn:${{{AWS_PARTITION}}}:ssm:${{{AWS_REGION}}}:${{{AWS_ACCOUNT_ID}}}"
                            f":parameter${{{cw_config_parameter.title}}}"
                        ),
                    ],
                },
            ],
        },
    )
    return ecs_sd_policy


def add_cw_agent_to_family(family, **options):
    """
    Function to add the CW Agent to the task family for additional monitoring
    :param ecs_composex.common.compose_services.ComposeFamily family:
    """
    prometheus_config = set_cw_prometheus_config_parameter(family)
    cw_agent_config = set_cw_config_parameter(family, **options)
    family.add_service(
        define_cloudwatch_agent(family, prometheus_config, cw_agent_config)
    )
    family.refresh()
    task_role = family.template.resources[ecs_params.TASK_ROLE_T]
    exec_role = family.template.resources[ecs_params.EXEC_ROLE_T]
    ecs_sd_policy = set_ecs_cw_policy(prometheus_config, cw_agent_config)
    if hasattr(task_role, "Policies") and isinstance(
        getattr(task_role, "Policies"), list
    ):
        policies = getattr(task_role, "Policies")
        policies.append(ecs_sd_policy)
    elif not hasattr(task_role, "Policies"):
        setattr(task_role, "Policies", [ecs_sd_policy])
    if hasattr(exec_role, "Policies") and isinstance(
        getattr(exec_role, "Policies"), list
    ):
        policies = getattr(exec_role, "Policies")
        policies.append(ecs_sd_policy)
    elif not hasattr(exec_role, "Policies"):
        setattr(exec_role, "Policies", [ecs_sd_policy])



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to define CloudWatch alarms in x-alarms and match these to services
"""

from compose_x_common.compose_x_common import keyisset
from troposphere import GetAtt, Output, Ref
from troposphere.cloudwatch import Alarm as CWAlarm
from troposphere.cloudwatch import CompositeAlarm, MetricDimension

from ecs_composex.alarms.alarms_stack import Alarm, create_alarms
from ecs_composex.common import LOG, add_outputs, add_parameters
from ecs_composex.common.cfn_params import Parameter
from ecs_composex.ecs.ecs_params import CLUSTER_NAME, SERVICE_SCALING_TARGET
from ecs_composex.ecs.ecs_scaling import (
    generate_alarm_scaling_out_policy,
    reset_to_zero_policy,
)
from ecs_composex.sns.sns_params import RES_KEY as SNS_KEY
from ecs_composex.sns.sns_params import TOPIC_ARN, TOPIC_ARN_RE
from ecs_composex.sns.sns_stack import Topic


def get_alarm_actions(alarm):
    """
    Function to get the alarm actions

    :param alarm:
    :return: the okay and alarm actions
    :rtype: tuple
    """
    if hasattr(alarm.cfn_resource, "OKActions"):
        okay_actions = getattr(alarm.cfn_resource, "OKActions")
    else:
        okay_actions = []
        setattr(alarm.cfn_resource, "OKActions", okay_actions)
    if hasattr(alarm.cfn_resource, "AlarmActions"):
        alarm_actions = getattr(alarm.cfn_resource, "AlarmActions")
    else:
        alarm_actions = []
        setattr(alarm.cfn_resource, "AlarmActions", alarm_actions)
    return okay_actions, alarm_actions


def handle_services_alarm_access(alarm):
    """
    Function to grant describe access onto the alarm

    :param alarm:
    :return:
    """


def add_service_actions(
    alarm, alarms_stack, target, scaling_in_policy, scaling_out_policy
):
    """
    Function to update the alarm properties with OKActions and AlarmActions

    :param ecs_composex.alarms.alarms_stack.Alarm alarm:
    :param ecs_composex.common.stacks.ComposeXStack alarms_stack:
    :param tuple target:
    :param scaling_in_policy:
    :param scaling_out_policy:
    """
    setattr(
        alarm,
        "Threshold",
        float(
            scaling_out_policy.StepScalingPolicyConfiguration.StepAdjustments[
                0
            ].MetricIntervalLowerBound
        ),
    )
    if not alarm.cfn_resource:
        raise AttributeError(f"Alarm {alarm.logical_name} has no CFN object associated")
    service_scaling_in_policy_param = Parameter(
        f"{target[0].logical_name}ScaleInPolicy", Type="String"
    )
    service_scaling_out_policy_param = Parameter(
        f"{target[0].logical_name}ScaleOutPolicy", Type="String"
    )
    add_parameters(
        alarms_stack.stack_template,
        [service_scaling_in_policy_param, service_scaling_out_policy_param],
    )
    add_outputs(
        target[0].template,
        [
            Output(
                f"{target[0].logical_name}ScaleInPolicy",
                Value=Ref(scaling_in_policy),
            ),
            Output(
                f"{target[0].logical_name}ScaleOutPolicy",
                Value=Ref(scaling_out_policy),
            ),
        ],
    )
    alarms_stack.Parameters.update(
        {
            service_scaling_in_policy_param.title: GetAtt(
                target[0].logical_name,
                f"Outputs.{target[0].logical_name}ScaleInPolicy",
            ),
            service_scaling_out_policy_param.title: GetAtt(
                target[0].logical_name,
                f"Outputs.{target[0].logical_name}ScaleOutPolicy",
            ),
        }
    )
    actions = get_alarm_actions(alarm)
    actions[0].append(Ref(service_scaling_in_policy_param))
    actions[1].append(Ref(service_scaling_out_policy_param))


def handle_service_scaling(alarm, alarms_stack):
    """
    Function to create the scaling steps for defined services

    :param ecs_composex.alarms.alarms_stack.Alarm alarm:
    :param ecs_composex.common.stacks.ComposeXStack alarms_stack:
    """
    for target in alarm.families_scaling:
        if SERVICE_SCALING_TARGET not in target[0].template.resources:
            LOG.warning(
                f"No Scalable target defined for {target[0].name}."
                " You need to define `scaling.scaling_range` in x-configs first. No scaling applied"
            )
            return
        scaling_out_policy = generate_alarm_scaling_out_policy(
            target[0].logical_name,
            target[0].template,
            target[1],
            scaling_source=alarm.logical_name,
        )
        scaling_in_policy = reset_to_zero_policy(
            target[0].logical_name,
            target[0].template,
            target[1],
            scaling_source=alarm.logical_name,
        )
        add_service_actions(
            alarm, alarms_stack, target, scaling_in_policy, scaling_out_policy
        )


def map_topic_to_action(alarm, notify_on, topic_identifier):
    """
    Function to map the topic to specific actions

    :param alarm: alarm props for alarms
    :param str notify_on:
    :param topic_identifier:
    :return:
    """
    actions = get_alarm_actions(alarm)
    if notify_on == "all":
        actions[0].append(topic_identifier)
        actions[1].append(topic_identifier)
    elif notify_on == "alarm":
        actions[1].append(topic_identifier)
    elif notify_on == "okay":
        actions[0].append(topic_identifier)


def handle_notify_on(topic_def):
    """
    Function to validate parameter NotifyOn

    :param dict topic_def:
    :return:
    """
    valid_values = ["all", "okay", "alarm"]
    notify_on = "alarm"
    if not keyisset("NotifyOn", topic_def):
        LOG.warning("NotifyOn was not set for topic. Will default to AlarmActions")
    else:
        if not isinstance(topic_def["NotifyOn"], str):
            raise TypeError("NotifyOn must be a string")
        notify_on = topic_def["NotifyOn"].lower()
        if notify_on not in valid_values:
            raise ValueError(
                "The value for NotifyOn",
                notify_on,
                "Is not valid. Expected one of",
                valid_values,
            )
    return notify_on


def handle_compose_topics(alarm, alarms_stack, settings, topic_def, notify_on):
    """
    Function to handle x-alarms to x-sns topics

    :param ecs_composex.alarms.alarms_stack.Alarm alarm:
    :param ecs_composex.common.stacks.ComposeXStack alarms_stack:
    :param ecs_composex.common.settings.ComposeXSettings settings:
    :param dict topic_def:
    :param str notify_on:
    :return:
    """
    try:
        topic = settings.compose_content[SNS_KEY][Topic.keyword][topic_def[SNS_KEY]]
    except KeyError:
        LOG.error(f"No topic {topic_def[SNS_KEY]} found in {SNS_KEY}.{Topic.keyword}")
        raise
    if not topic.attributes_outputs:
        topic.init_outputs()
        topic.generate_outputs()
    topic_arn = topic.attributes_outputs[TOPIC_ARN]
    if topic.cfn_resource:
        add_parameters(alarms_stack.stack_template, [topic_arn["ImportParameter"]])
        alarms_stack.Parameters.update({topic_arn["Name"]: topic_arn["ImportValue"]})
        map_topic_to_action(alarm, notify_on, Ref(topic_arn["ImportParameter"]))
    else:
        if keyisset(SNS_KEY, settings.mappings) and keyisset(
            topic.logical_name, settings.mappings[SNS_KEY]
        ):
            alarms_stack.stack_template.add_mapping(
                topic.module_name, settings.mappings[SNS_KEY]
            )
            map_topic_to_action(alarm, notify_on, topic_arn["ImportValue"])


def handle_alarm_topics(alarm, alarms_stack, settings):
    """
    Function to add the topics actions for defined topics

    :param ecs_composex.alarms.alarms_stack.Alarm alarm:
    :param ecs_composex.common.stacks.ComposeXStack alarms_stack:
    :param ecs_composex.common.settings.ComposeXSettings settings:
    :return:
    """

    for topic in alarm.topics:
        notify_on = handle_notify_on(topic)
        if keyisset("TopicArn", topic):
            if not TOPIC_ARN_RE.match(topic["TopicArn"]):
                raise ValueError("Invalid ARN for topic", topic["TopicArn"])
            map_topic_to_action(alarm, notify_on, topic["TopicArn"])
        elif keyisset(SNS_KEY, topic):
            if not keyisset(SNS_KEY, settings.compose_content) or (
                keyisset(SNS_KEY, settings.compose_content)
                and not keyisset(
                    topic[SNS_KEY],
                    settings.compose_content[SNS_KEY][Topic.keyword],
                )
            ):
                raise KeyError(
                    f"There is no topic {topic[SNS_KEY]} defined in {SNS_KEY}"
                )
            handle_compose_topics(alarm, alarms_stack, settings, topic, notify_on)


def alarms_to_ecs(resources, services_stack, res_root_stack, settings):
    new_resources = [
        resource
        for resource in resources.values()
        if not resource.lookup and not resource.use
    ]
    for alarm in new_resources:
        if alarm.in_composite:
            continue
        handle_service_scaling(alarm, res_root_stack)
        if alarm.topics:
            handle_alarm_topics(alarm, res_root_stack, settings)


def update_alarm_threshold(alarm_properties, settings):
    """
    Function to update the threshold based on the defined settings

    :param alarm_properties:
    :param settings:
    :return:
    """
    if alarm_properties["MetricName"] in settings and keyisset(
        alarm_properties["MetricName"], settings
    ):
        alarm_properties["Threshold"] = float(settings[alarm_properties["MetricName"]])


def update_definition_from_settings(alarm_definition, settings):
    """
    Function to update the alarm definition with the global settings

    :param dict alarm_definition:
    :param dict settings:
    :return:
    """
    alarm_definition["Properties"].update(
        {
            "DatapointsToAlarm": settings["DatapointsToAlarm"],
            "EvaluationPeriods": settings["EvaluationPeriods"],
            "Period": settings["Period"],
        }
    )


def set_services_alarms(settings):
    """
    Function to create and assign alarms to services

    :param ecs_composex.common.settings.ComposeXSettings settings:
    :return:
    """
    for family in settings.families.values():
        if not family.predefined_alarms:
            continue
        family_alarms = []
        for name, definition in family.predefined_alarms.items():
            primary_name = definition["Primary"]
            primary = definition["Alarms"][primary_name]
            update_definition_from_settings(primary, definition["Settings"])
            for alarm_name, alarm_definition in definition["Alarms"].items():
                if keyisset("Topics", definition):
                    alarm_definition["Topics"] = definition["Topics"]
                if keyisset("Properties", alarm_definition) and keyisset(
                    "MetricName", alarm_definition["Properties"]
                ):
                    update_alarm_threshold(
                        alarm_definition["Properties"], definition["Settings"]
                    )
                the_alarm = Alarm(
                    alarm_name, alarm_definition, family.logical_name, settings
                )
                family_alarms.append(the_alarm)
        create_alarms(family.template, settings, family_alarms)
        for alarm in family_alarms:
            dimensions = [
                MetricDimension(**{"Name": "ClusterName", "Value": Ref(CLUSTER_NAME)}),
                MetricDimension(
                    **{
                        "Name": "ServiceName",
                        "Value": GetAtt(family.ecs_service.ecs_service, "Name"),
                    }
                ),
            ]
            if isinstance(alarm.cfn_resource, CWAlarm):
                setattr(alarm.cfn_resource, "Dimensions", dimensions)
            if issubclass(type(alarm.cfn_resource), CompositeAlarm):
                handle_alarm_topics(alarm, family.stack, settings)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from os import path

from ecs_composex.common.cfn_params import Parameter
from ecs_composex.common.ecs_composex import X_KEY

MOD_KEY = path.basename(path.dirname(path.abspath(__file__)))
RES_KEY = f"{X_KEY}{MOD_KEY}"

S3_ARN_REGEX = r"arn:(aws|aws-gov|aws-cn):s3:::([a-zA-Z0-9-.]+$)"

S3_BUCKET_ARN_T = "BucketArn"
S3_BUCKET_ARN = Parameter(
    S3_BUCKET_ARN_T,
    return_value="Arn",
    Type="String",
    AllowedPattern=S3_ARN_REGEX,
)
S3_BUCKET_NAME_T = "BucketName"
S3_BUCKET_NAME = Parameter(
    S3_BUCKET_NAME_T, Type="String", AllowedPattern=r"^[a-z0-9-.]+$"
)
S3_BUCKET_DOMAIN_NAME_T = "BucketDomainName"
S3_BUCKET_DOMAIN_NAME = Parameter(
    S3_BUCKET_DOMAIN_NAME_T, return_value="DomainName", Type="String"
)

S3_BUCKET_REGION_DOMAIN_NAME_T = "BucketDomainName"
S3_BUCKET_REGION_DOMAIN_NAME = Parameter(
    S3_BUCKET_REGION_DOMAIN_NAME_T,
    return_value="RegionalDomainName",
    Type="String",
)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from troposphere import Equals, Not, Ref

from ecs_composex.dns import dns_params

CREATE_PUBLIC_NAMESPACE_CON_T = "CreatePublicServicesNamespaceCondition"
CREATE_PUBLIC_NAMESPACE_CON = Not(
    Equals(
        Ref(dns_params.PUBLIC_DNS_ZONE_NAME),
        dns_params.PUBLIC_DNS_ZONE_NAME.Default,
    )
)
CREATE_PUBLIC_ZONE_CON_T = "CreatePublicServicesZoneCondition"
CREATE_PUBLIC_ZONE_CON = Not(
    Equals(
        Ref(dns_params.PUBLIC_DNS_ZONE_NAME),
        dns_params.PUBLIC_DNS_ZONE_NAME.Default,
    )
)
CREATE_PRIVATE_NAMESPACE_CON_T = "CreatePrivateServicesNamespaceCondition"
CREATE_PRIVATE_NAMESPACE_CON = Equals(
    Ref(dns_params.PRIVATE_DNS_ZONE_ID), dns_params.PRIVATE_DNS_ZONE_ID.Default
)

USE_DEFAULT_ZONE_NAME_CON_T = "UseDefaultPrivateZoneName"
USE_DEFAULT_ZONE_NAME_CON = Equals(
    Ref(dns_params.PRIVATE_DNS_ZONE_NAME),
    dns_params.PRIVATE_DNS_ZONE_NAME.Default,
)

PRIVATE_ZONE_ID_CON_T = "PrivateNamespaceZoneIdCondition"
PRIVATE_ZONE_ID_CON = Not(
    Equals(
        Ref(dns_params.PRIVATE_DNS_ZONE_ID),
        dns_params.PRIVATE_DNS_ZONE_ID.Default,
    )
)

PRIVATE_NAMESPACE_CON_T = "PrivateNamespaceIdCondition"
PRIVATE_NAMESPACE_CON = Not(
    Equals(
        Ref(dns_params.PRIVATE_NAMESPACE_ID),
        dns_params.PRIVATE_NAMESPACE_ID.Default,
    )
)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from os import path

from ecs_composex.common.cfn_params import Parameter
from ecs_composex.common.ecs_composex import X_KEY

TABLE_NAME_T = "TableName"
TABLE_NAME = Parameter(TABLE_NAME_T, Type="String", AllowedPattern=r"[a-zA-Z0-9_.-]+")

TABLE_ARN_T = "Arn"
TABLE_ARN = Parameter(TABLE_ARN_T, return_value="Arn", Type="String")

MOD_KEY = path.basename(path.dirname(path.abspath(__file__)))
RES_KEY = f"{X_KEY}{MOD_KEY}"

TABLE_SSM_PREFIX = f"/{RES_KEY}/"



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Most commonly used functions shared across all modules.
"""


import logging as logthings
import re
import sys
from datetime import datetime as dt
from os import environ
from uuid import uuid4

import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from compose_x_common.compose_x_common import keyisset, keypresent
from troposphere import AWS_NO_VALUE, Output, Parameter, Ref, Template

from ecs_composex import __version__ as version
from ecs_composex.common import cfn_conditions, cfn_params

DATE = dt.utcnow().isoformat()
FILE_PREFIX = f'{dt.utcnow().strftime("%Y/%m/%d/%H%M")}/{str(uuid4().hex)[:6]}'
NONALPHANUM = re.compile(r"([^a-zA-Z0-9]+)")

EXIT_CODES = {"MODULE_NOT_FOUND": 8, "MISSING_RESOURCE_DEFINITION": 9}


def no_value_if_not_set(props, key, is_bool=False):
    """
    Function to simplify setting value if the key is in the dict and else Ref(AWS_NO_VALUE) for resource properties

    :param dict props:
    :param str key:
    :param bool is_bool:
    :return:
    """
    if not is_bool:
        return Ref(AWS_NO_VALUE) if not keyisset(key, props) else props[key]
    else:
        return Ref(AWS_NO_VALUE) if not keypresent(key, props) else props[key]


def init_template(description=None):
    """Function to initialize the troposphere base template

    :param description: Description used for the CFN
    :type description: str

    :returns: template
    :rtype: Template
    """
    if description is not None:
        template = Template(description)
    else:
        template = Template("Template generated by ECS ComposeX")
    template.set_metadata(
        {
            "Type": "ComposeX",
            "Properties": {"Version": version, "GeneratedOn": DATE},
        }
    )
    template.set_version()
    return template


def add_parameters(template, parameters):
    """Function to add parameters to the template

    :param template: the template to add the parameters to
    :type template: troposphere.Template
    :param parameters: list of parameters to add to the template
    :type parameters: list<troposphere.Parmeter>
    """
    for param in parameters:
        if not isinstance(param, Parameter):
            raise TypeError("Parameter must be of type", Parameter)
        if template and param.title not in template.parameters:
            template.add_parameter(param)


def add_outputs(template, outputs):
    """Function to add parameters to the template

    :param template: the template to add the parameters to
    :type template: troposphere.Template
    :param outputs: list of parameters to add to the template
    :type outputs: list<troposphere.Output>
    """
    for output in outputs:
        if not isinstance(output, Output):
            raise TypeError("Parameter must be of type", Output)
        if template and output.title not in template.outputs:
            template.add_output(output)


def add_defaults(template):
    """Function to CFN parameters and conditions to the template whhich are used
    across ECS ComposeX

    :param template: source template to add the params and conditions to
    :type template: Template
    """
    template.add_parameter(cfn_params.ROOT_STACK_NAME)
    template.add_condition(
        cfn_conditions.USE_STACK_NAME_CON_T, cfn_conditions.USE_STACK_NAME_CON
    )


def build_template(description=None, *parameters):
    """
    Entry point function to creating the template for ECS ComposeX resources

    :param description: Optional custom description for the CFN template
    :type description: str, optional
    :param parameters: List of optional parameters to add to the template.
    :type parameters: List<troposphere.Parameters>, optional

    :returns template: the troposphere template
    :rtype: Template
    """
    template = init_template(description)
    if parameters:
        add_parameters(template, *parameters)
    add_defaults(template)
    return template


def validate_resource_title(resource_name, resource_type=None):
    """Function to validate the key for the resource is valid

    :param resource_name: Name of the resource to evaluate
    :type resource_name: str
    :param resource_type: category of the resource, optional
    :type resource_type: str

    :returns: True/False
    :rtype: bool
    """
    if NONALPHANUM.findall(resource_name):
        raise ValueError(
            f"The resource {resource_name} in {resource_type} "
            "section contains non alphanumerical characters",
            NONALPHANUM.findall(resource_name),
        )
    return True


def validate_input(compose_content, res_key):
    """Function to validate the resources names in ComposeX File
    for a given resource key

    :param compose_content: the docker/ComposeX content
    :type compose_content: dict
    :param res_key: key of the category in docker compose to look for
    :type res_key: str

    :return: True/False if all keys are valid
    :rtype: bool
    """
    section = compose_content[res_key]
    for resource_name in section.keys():
        validate_resource_title(resource_name, res_key)
    return True


def validate_kwargs(required_keys, kwargs, caller=None):
    """Function to ensure minimum keys in kwargs are present.

    :param required_keys: the list of keys that have to be present
    :type required_keys: list
    :param kwargs: the arguments to verify
    :type kwargs: dict or set
    :param caller: optional argument to help understand what's missing
    :type caller: str

    :return: True/False if all keys are valid
    :rtype: bool

    :raises: KeyError if key is missing from kwargs
    """
    for required_key in required_keys:
        if not keyisset(required_key, kwargs):
            raise KeyError(required_key, "is required by module", caller)
    return True


def setup_logging():
    """Function to setup logging for ECS ComposeX.
    In case this is used in a Lambda function, removes the AWS Lambda default log handler

    :returns: the_logger
    :rtype: Logger
    """
    level = environ.get("LOGLEVEL")
    default_level = True
    formats = {
        "INFO": logthings.Formatter(
            "%(asctime)s [%(levelname)s], %(message)s",
            "%Y-%m-%d %H:%M:%S",
        ),
        "DEBUG": logthings.Formatter(
            "%(asctime)s [%(levelname)s], %(filename)s.%(lineno)d , %(funcName)s, %(message)s",
            "%Y-%m-%d %H:%M:%S",
        ),
    }

    if level is not None and isinstance(level, str):
        print("SETTING TO", level.upper())
        logthings.basicConfig(level=level.upper())
        default_level = False
    else:
        logthings.basicConfig(level="INFO")

    root_logger = logthings.getLogger()
    for h in root_logger.handlers:
        root_logger.removeHandler(h)
    the_logger = logthings.getLogger("EcsComposeX")

    if not the_logger.handlers:
        if default_level:
            formatter = formats["INFO"]
        elif keyisset(level.upper(), formats):
            formatter = formats[level.upper()]
        else:
            formatter = formats["DEBUG"]
        handler = logthings.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        the_logger.addHandler(handler)

    return the_logger


def build_parameters_file(params, parameter_name, parameter_value):
    """
    Function to build arguments file to pass onto CFN.
    Adds the parameter key/value so it can be written to file afterwards

    :param params: list of parameters
    :type params: list
    :param parameter_name: key of the parameter
    :type parameter_name: str
    :param parameter_value: value of the parameter
    :type parameter_value: str||list
    """
    if params is None:
        params = []
    if isinstance(parameter_value, (int, float)):
        parameter_value = str(parameter_value)
    params.append({"ParameterKey": parameter_name, "ParameterValue": parameter_value})


def build_default_stack_parameters(stack_params, **kwargs):
    """
    Function to check and define default parameters for the root stack from the CLI options

    :param stack_params: list of parameters to add to to use for the root stack
    :type stack_params: list
    :param kwargs: extended arguments
    :type kwargs: dict
    """
    if keyisset(cfn_params.USE_FLEET_T, kwargs):
        build_parameters_file(
            stack_params, cfn_params.USE_FLEET_T, kwargs[cfn_params.USE_FLEET_T]
        )


def load_composex_file(file_path):
    """
    File to load and read the docker compose file

    :param file_path: path to the docker compose file
    :type file_path: str

    :return: content of the docker file
    :rtype: dict
    """
    with open(file_path, "r") as composex_fd:
        return yaml.load(composex_fd.read(), Loader=Loader)


LOG = setup_logging()



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>


"""
Module to handle RDS Features definition
"""

from compose_x_common.compose_x_common import keyisset
from troposphere import AWS_NO_VALUE, AWS_STACK_NAME, GetAtt, Ref, Sub
from troposphere.iam import Role as IamRole
from troposphere.rds import DBClusterRole

from ecs_composex.common import LOG
from ecs_composex.iam import define_iam_policy, service_role_trust_policy
from ecs_composex.rds.rds_features_define import (
    define_s3_export_feature_policy,
    define_s3_import_feature_policy,
)

S3_KEY = "x-s3"


def validate_rds_features(features_list, db_family):
    """
    Function to validate the features listed are supported by the given family.

    :param features_list:
    :param db_family:
    :return:
    """


def validate_features_content(data):
    """
    Function to ensure the data given in compose file is valid for S3/MacroParameters/IamAccess

    :param dict allowed_keys:
    :param dict data:
    :return:
    """
    feature_structure = {"Name": (str, None), "Resources": (list, None)}
    for feature in data:
        for name, rtype in feature_structure.items():
            if not keyisset(name, feature):
                raise KeyError(f"Features requires {name}. Got", feature.keys())
            elif not isinstance(feature[name], rtype[0]):
                raise TypeError(
                    f"Feature property {name} is of type",
                    type(feature),
                    "Expected",
                    rtype[0],
                )


def define_associated_roles(db):
    """
    Function to define the AssociatedRoles, either present or empty
    :param ecs_composex.rds.rds_stack.RdsDb db:
    :return: the list of Associated Roles
    :rtype: list
    """
    if db.cfn_resource and hasattr(db.cfn_resource, "AssociatedRoles"):
        LOG.warning(
            "The db properties already had AssociatedRoles defined."
            " Only will add ones without the feature already defined"
        )
        roles = getattr(db.cfn_resource, "AssociatedRoles")
    else:
        roles = []
    return roles


def add_rds_features(settings, stack, db, features, db_template, boundary):
    """
    Function to add AssociatedRoles and Features if not already defined in the DB properties for that feature.
    """
    features_settings = {
        "s3Import": define_s3_import_feature_policy,
        "s3Export": define_s3_export_feature_policy,
        "Lambda": None,
        "SageMaker": None,
        "Comprehend": None,
    }
    validate_features_content(features)
    roles = define_associated_roles(db)
    to_add = [
        feature
        for feature in features
        if feature["Name"] not in [role.FeatureName for role in roles]
    ]
    excluded = [
        feature
        for feature in features
        if feature["Name"] in [role.FeatureName for role in roles]
    ]
    if excluded:
        LOG.warning(
            f"Features {excluded} are not being processed as already defined in AssociatedRoles of the DB properties"
        )
    if not to_add:
        LOG.warning("No features were found to be added at all!!")
        return
    policies = []
    iam_role = IamRole(
        f"{db.logical_name}FeaturesIamRole",
        AssumeRolePolicyDocument=service_role_trust_policy("rds"),
        Description=Sub(
            f"{db.logical_name} RDS Features IAM Role in ${{{AWS_STACK_NAME}}}"
        ),
        Policies=policies,
        PermissionsBoundary=boundary,
        MaxSessionDuration=3600,
    )
    features_definition = []
    for feature in to_add:
        if feature["Name"] not in features_settings.keys() or not (
            feature["Name"] in features_settings.keys()
            and features_settings[feature["Name"]]
        ):
            LOG.warning(
                f"The feature {feature['Name']} is not currently supported. Sorry."
            )
        policies.append(
            features_settings[feature["Name"]](
                settings, stack, db, feature["Resources"], db_template
            )
        )
        features_definition.append(
            DBClusterRole(FeatureName=feature["Name"], RoleArn=GetAtt(iam_role, "Arn"))
        )
    db_template.add_resource(iam_role)
    if policies and not hasattr(db.cfn_resource, "AssociatedRoles"):
        setattr(db.cfn_resource, "AssociatedRoles", features_definition)


def apply_extra_parameters(settings, stack, db, db_template):
    """
    Function to add extra parameters set in MacroParameters post creation of the DB resource from properties

    :param ecs_composex.rds.rds_stack.Rds db db:
    :param troposphere.Template db_template:
    :return:
    """
    if not db.parameters:
        return
    permissions_boundary = Ref(AWS_NO_VALUE)
    if keyisset("PermissionsBoundary", db.parameters):
        permissions_boundary = define_iam_policy(db.parameters["PermissionsBoundary"])
    extra_parameters = {"RdsFeatures": (list, add_rds_features)}
    for name, config in extra_parameters.items():
        if not keyisset(name, db.parameters):
            LOG.debug(
                f"Feature {name} has not been set in compose file. {db.parameters}"
            )
        if (
            keyisset(name, db.parameters)
            and isinstance(db.parameters[name], config[0])
            and config[1]
        ):
            config[1](
                settings,
                stack,
                db,
                db.parameters[name],
                db_template,
                permissions_boundary,
            )
        elif keyisset(name, db.parameters) and not isinstance(
            db.parameters[name], config[0]
        ):
            LOG.error(
                f"The property {name} is of type {type(db.parameters[name])}. Expected {config[0]}. Skipping"
            )



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from troposphere import Tags
from troposphere.kinesis import Stream, StreamEncryption

from ecs_composex.common import LOG, build_template, keyisset
from ecs_composex.resources_import import import_record_properties


def handle_encryption(stream):
    """
    Function to define the encryption

    :param stream:
    :return:
    """
    return StreamEncryption(
        EncryptionType="KMS",
        KeyId=stream.properties["StreamEncryption"]["KeyId"],
    )


def create_new_stream(stream):
    """
    Function to create the new Kinesis stream
    :param ecs_composex.kinesis.kinesis_stack.Stream stream:
    :param ecs_composex.common.settings.ComposeXSettings settings:
    :return:
    """
    props = import_record_properties(stream.properties, Stream)
    if not keyisset("ShardCount", stream.properties):
        LOG.warning("ShardCount must be set. Defaulting to 1")
        props["ShardCount"] = 1
    props["Tags"] = Tags(Name=stream.logical_name, ComposeName=stream.name)
    stream.cfn_resource = Stream(stream.logical_name, **props)
    stream.init_outputs()
    stream.generate_outputs()


def create_streams_template(new_resources, settings):
    """
    Function to create the root template for Kinesis streams

    :param list<ecs_composex.kinesis.kinesis_stack.Stream> new_resources:
    :param ecs_composex.common.settings.ComposeXSettings settings:
    :return:
    """
    root_template = build_template("Root stack for ecs_composex.kinesis")
    for res in new_resources:
        create_new_stream(res)
        root_template.add_resource(res.cfn_resource)
        root_template.add_output(res.outputs)
    return root_template



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
AppMesh related conditions
"""

from troposphere import AWS_ACCOUNT_ID, AWS_STACK_NAME, Equals, GetAtt, If, Ref
from troposphere.appmesh import Mesh

from ecs_composex.appmesh.appmesh_params import MESH_NAME, MESH_OWNER_ID

USER_IS_SELF_CON_T = "AppMeshOwnerIsSelf"
USER_IS_SELF_CON = Equals(Ref(MESH_OWNER_ID), MESH_OWNER_ID.Default)

USE_DEFAULT_MESH_NAME_CON_T = "UseDefaultMeshNameCondition"
USE_DEFAULT_MESH_NAME_CON = Equals(Ref(MESH_NAME), MESH_NAME.Default)


def set_mesh_owner_id():
    """
    Returns the IF condition for MeshOwner

    :return: CFN Condition
    :rtype: If
    """
    return If(USER_IS_SELF_CON_T, Ref(AWS_ACCOUNT_ID), Ref(MESH_OWNER_ID))


def set_mesh_name():
    """
    Function to set the mesh name if default is passed, otherwise generate it.
    :return: If
    """
    return If(USE_DEFAULT_MESH_NAME_CON_T, Ref(AWS_STACK_NAME), Ref(MESH_NAME))


def get_mesh_name(mesh_obj):
    """
    Function to return the mesh reference based on the type of parameter mesh_obj

    :return:
    """
    if isinstance(mesh_obj, Mesh):
        return GetAtt(mesh_obj, "MeshName")
    elif isinstance(mesh_obj, Ref):
        return Ref(MESH_NAME)


def get_mesh_owner(mesh_obj):
    """
    Function to return the mesh reference based on the type of parameter mesh_obj

    :return:
    """
    if isinstance(mesh_obj, Mesh):
        return GetAtt(mesh_obj, "MeshOwner")
    return Ref(MESH_OWNER_ID)


def add_appmesh_conditions(template):
    """
    Function to add AppMesh default conditions to template

    :param troposphere.Template template: The template to add the conditions to
    """
    template.add_condition(USE_DEFAULT_MESH_NAME_CON_T, USE_DEFAULT_MESH_NAME_CON)
    template.add_condition(USER_IS_SELF_CON_T, USER_IS_SELF_CON)



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from ecs_composex import __version__ as version

metadata = {
    "Type": "ComposeX",
    "Properties": {
        "ecs_composex::module": "ecs_composex.kms",
        "Version": version,
    },
}



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>


from compose_x_common.compose_x_common import keyisset

from ecs_composex.common import LOG


class ComposeXConfig(object):
    """
    Class to parse and define configuration settings for ECS ComposeX
    """

    master_key = "x-configs"
    composex_key = "composex"

    def __init__(self, settings):
        """
        Initializes the ComposeXConfig class

        :param ComposeXSettings settings: The execution settings
        """
        self.composex_config = {}
        if keyisset(self.master_key, settings.compose_content) and keyisset(
            self.composex_key, settings.compose_content[self.master_key]
        ):
            self.composex_config = settings.compose_content[self.master_key][
                self.composex_key
            ]


class ComputeConfig(ComposeXConfig):
    """
    Class to determine the compute settings to use when deploying on top of EC2.
    """

    default_spot_config = {
        "use_spot": True,
        "bid_price": 0.42,
        "spot_instance_types": {
            "m5a.xlarge": {"weight": 3},
            "m5a.2xlarge": {"weight": 7},
            "m5a.4xlarge": {"weight": 15},
        },
    }
    spot_key = "spot_config"

    def __init__(self, settings):
        """
        Method to initialize Compute config

        :param ecs_composex.common.settings.ComposeXSettings settings: The settings for execution
        """
        super().__init__(settings)
        if keyisset(self.spot_key, self.composex_config):
            self.spot_config = self.composex_config[self.spot_key]
        else:
            LOG.warning(
                "No spot_config set in configs of ComposeX File. Setting to defaults"
            )
            self.spot_config = self.default_spot_config



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to find the CodeGuru profiles from tagging API
"""

import re

from ecs_composex.codeguru_profiler.codeguru_profiler_params import (
    PROFILER_ARN,
    PROFILER_NAME,
)
from ecs_composex.common import LOG
from ecs_composex.common.aws import (
    define_lookup_role_from_info,
    find_aws_resource_arn_from_tags_api,
)


def get_profile_config(profile_arn, session):
    """

    :param str profile_arn:
    :param boto3.session.Session session:
    :return:
    """
    profile_parts = re.compile(
        r"(?:^arn:aws(?:-[a-z]+)?:codeguru-profiler:[\S]+:[0-9]{12}:profilingGroup/)([\S]+)$"
    )
    profile_name = profile_parts.match(profile_arn).groups()[0]
    profile_config = {
        PROFILER_ARN.title: profile_arn,
        PROFILER_NAME.title: profile_name,
    }
    return profile_config


def lookup_profile_config(lookup, session):
    """
    Function to find the DB in AWS account

    :param dict lookup: The Lookup definition for DB
    :param boto3.session.Session session: Boto3 session for clients
    :return:
    """
    lookup_types = {
        "codeguru-profiler": {
            "regexp": r"(?:^arn:aws(?:-[a-z]+)?:codeguru-profiler:[\S]+:[0-9]{12}:profilingGroup/)([\S]+)$"
        },
    }
    lookup_session = define_lookup_role_from_info(lookup, session)
    profile_arn = find_aws_resource_arn_from_tags_api(
        lookup,
        lookup_session,
        "codeguru-profiler",
        types=lookup_types,
    )
    if not profile_arn:
        return None
    config = get_profile_config(profile_arn, lookup_session)
    LOG.debug(config)
    return config



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Main module generating the ECS Cluster template.

The root stack is to build the IAM Instance profile for the hosts that can be used for ASG or SpotFleet.
That way it is easy for anyone to deploy an instance in standalone if you wanted that.
"""

from troposphere import GetAtt, If, Ref

from ecs_composex.common import build_template, cfn_conditions
from ecs_composex.common.cfn_params import (
    ROOT_STACK_NAME,
    ROOT_STACK_NAME_T,
    USE_FLEET,
    USE_ONDEMAND,
)
from ecs_composex.common.config import ComputeConfig
from ecs_composex.common.stacks import ComposeXStack
from ecs_composex.compute import compute_conditions, compute_params
from ecs_composex.compute.hosts_template import add_hosts_resources
from ecs_composex.compute.spot_fleet import generate_spot_fleet_template
from ecs_composex.ecs.ecs_params import CLUSTER_NAME
from ecs_composex.vpc import vpc_params


def add_spotfleet_stack(template, settings, launch_template):
    """
    Function to build the spotfleet stack and add it to the Cluster parent template

    :param launch_template: the launch template
    :param troposphere.Template template: parent cluster template
    :param ComposeXSettings settings: The settings for execution

    """
    compute_config = ComputeConfig(settings)
    parameters = {
        ROOT_STACK_NAME_T: If(
            cfn_conditions.USE_STACK_NAME_CON_T,
            Ref("AWS::StackName"),
            Ref(ROOT_STACK_NAME),
        ),
        compute_params.LAUNCH_TEMPLATE_ID_T: Ref(launch_template),
        compute_params.LAUNCH_TEMPLATE_VersionNumber_T: GetAtt(
            launch_template, "LatestVersionNumber"
        ),
        compute_params.MAX_CAPACITY_T: Ref(compute_params.MAX_CAPACITY),
        compute_params.MIN_CAPACITY_T: Ref(compute_params.MIN_CAPACITY),
        compute_params.TARGET_CAPACITY_T: Ref(compute_params.TARGET_CAPACITY),
    }
    fleet_template = generate_spot_fleet_template(settings, compute_config.spot_config)
    template.add_resource(
        ComposeXStack(
            "SpotFleet",
            stack_template=fleet_template,
            Condition=cfn_conditions.USE_SPOT_CON_T,
            Parameters=parameters,
        )
    )


def generate_compute_template(settings):
    """
    Function that generates the Compute resources to run ECS services on top of EC2

    :param ComposeXSettings settings: The settings for executio
    :return: ECS Cluster Template
    :rtype: troposphere.Template
    """
    template = build_template(
        "Cluster template generated by ECS Compose X",
        [
            USE_FLEET,
            USE_ONDEMAND,
            compute_params.ECS_AMI_ID,
            compute_params.TARGET_CAPACITY,
            compute_params.MIN_CAPACITY,
            compute_params.MAX_CAPACITY,
            vpc_params.APP_SUBNETS,
            vpc_params.VPC_ID,
            CLUSTER_NAME,
        ],
    )
    template.add_condition(
        compute_conditions.MAX_IS_MIN_T, compute_conditions.MAX_IS_MIN
    )
    template.add_condition(cfn_conditions.USE_SPOT_CON_T, cfn_conditions.USE_SPOT_CON)
    launch_template = add_hosts_resources(template)
    add_spotfleet_stack(template, settings, launch_template)
    return template



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to help with defining the network settings for the ECS Service based on the family services definitions.
"""

from json import dumps

from compose_x_common.compose_x_common import keyisset, keypresent
from troposphere import AWS_ACCOUNT_ID, GetAtt, Parameter, Ref, Sub
from troposphere.ec2 import SecurityGroupIngress

from ecs_composex.common import LOG, add_parameters
from ecs_composex.ecs.ecs_params import SERVICE_NAME_T
from ecs_composex.ingress_settings import Ingress, set_service_ports
from ecs_composex.vpc.vpc_params import SG_ID_TYPE


def handle_ext_sources(existing_sources, new_sources):
    LOG.debug("Source", dumps(existing_sources, indent=2))
    set_ipv4_sources = [
        s[Ingress.ipv4_key] for s in existing_sources if keyisset(Ingress.ipv4_key, s)
    ]
    for new_s in new_sources:
        if new_s not in set_ipv4_sources:
            existing_sources.append(new_s)


def handle_aws_sources(existing_sources, new_sources):
    """
    Function to handle merge of aws sources between two services for one family
    :param existing_sources:
    :param new_sources:
    :return:
    """
    LOG.debug("Source", dumps(existing_sources, indent=2))
    set_ids = [s["Id"] for s in existing_sources if keyisset("Id", s)]
    allowed_keys = ["PrefixList", "SecurityGroup"]
    for new_s in new_sources:
        if new_s not in set_ids and new_s["Type"] in allowed_keys:
            existing_sources.append(new_s)
        elif new_s["Id"] not in allowed_keys:
            LOG.error(
                f"AWS Source type incorrect: {new_s['type']}. Expected one of {allowed_keys}"
            )


def handle_services(existing_sources, new_sources):
    """
    Function to merge source services definitions

    :param list existing_sources:
    :param list new_sources:
    :return:
    """
    set_ids = [s["Name"] for s in existing_sources if keyisset("Name", s)]
    for new_s in new_sources:
        if new_s not in set_ids:
            existing_sources.append(new_s)


def handle_ingress_rules(source_config, ingress_config):
    LOG.debug("Source", dumps(source_config, indent=2))
    valid_keys = [
        (ServiceNetworking.self_key, bool, None),
        (Ingress.ext_sources_key, list, handle_ext_sources),
        (Ingress.aws_sources_key, list, handle_aws_sources),
        (Ingress.services_key, list, handle_services),
    ]
    for key in valid_keys:
        if keypresent(key[0], ingress_config) and isinstance(
            ingress_config[key[0]], key[1]
        ):
            if key[1] is bool and not keyisset(key[0], source_config):
                source_config[key[0]] = ingress_config[key[0]]
            if (
                key[1] is bool
                and keyisset(key[0], source_config)
                and not keyisset(key[0], ingress_config)
            ):
                LOG.warning(
                    "At least one service in the task requires access to itself. Skipping."
                )
            elif key[1] is list and keyisset(key[0], ingress_config) and key[2]:
                key[2](source_config[key[0]], ingress_config[key[0]])


def handle_merge_services_props(config, network, network_config):
    """
    Function to handle properties assignment for network settings

    :param tuple config:
    :param dict network:
    :param dict network_config:
    :return:
    """
    if config[1] is bool and keypresent(config[0], network):
        network_config[config[0]] = network[config[0]]
    elif config[1] is str and keyisset(config[0], network):
        network_config[config[0]] = network[config[0]]
    elif config[1] is dict and keypresent(config[0], network) and config[2]:
        config[2](network_config[config[0]], network[config[0]])


def merge_services_network(family):
    network_config = {
        "UseCloudmap": False,
        Ingress.master_key: {
            ServiceNetworking.self_key: False,
            Ingress.ext_sources_key: [],
            Ingress.aws_sources_key: [],
            Ingress.services_key: [],
        },
        "IsPublic": False,
    }
    valid_keys = [
        (Ingress.master_key, dict, handle_ingress_rules),
        ("UseCloudmap", bool, None),
        ("IsPublic", bool, None),
    ]
    x_network = [s.x_network for s in family.ordered_services if s.x_network]
    for config in valid_keys:
        if config[1] is bool and any(
            [cfg[config[0]] for cfg in x_network if keypresent(config[0], cfg)]
        ):
            LOG.info(
                f"At least one service of {family.name} is set to use {config[0]}. Enabling."
            )
            network_config[config[0]] = True
        else:
            for network in x_network:
                if config[0] in network:
                    handle_merge_services_props(config, network, network_config)

    LOG.debug(family.name)
    LOG.debug(dumps(network_config, indent=2))
    return network_config


def add_independant_rules(dst_family, service_name, root_stack):
    src_service_stack = root_stack.stack_template.resources[service_name]
    for port in dst_family.service_config.network.ports:
        ingress_rule = SecurityGroupIngress(
            f"From{src_service_stack.title}To{dst_family.logical_name}On{port['published']}",
            FromPort=port["published"],
            ToPort=port["published"],
            IpProtocol=port["protocol"],
            Description=Sub(
                f"From {src_service_stack.title} to {dst_family.logical_name}"
                f" on port {port['published']}/{port['protocol']}"
            ),
            GroupId=GetAtt(
                dst_family.stack.title,
                f"Outputs.{dst_family.logical_name}GroupId",
            ),
            SourceSecurityGroupId=GetAtt(
                src_service_stack.title,
                f"Outputs.{src_service_stack.title}GroupId",
            ),
            SourceSecurityGroupOwnerId=Ref(AWS_ACCOUNT_ID),
        )
        if ingress_rule.title not in root_stack.stack_template.resources:
            root_stack.stack_template.add_resource(ingress_rule)


def set_compose_services_ingress(root_stack, dst_family, families, settings):
    """
    Function to crate SG Ingress between two families / services.
    Presently, the ingress rules are set after all services have been created

    :param ecs_composex.common.stacks.ComposeXStack root_stack:
    :param ecs_composex.common.compose_services.ComposeFamily dst_family:
    :param list families: The list of family names.
    :return:
    """
    for service in dst_family.service_config.network.services:
        service_name = service["Name"]
        if service_name not in families:
            raise KeyError(
                f"The service {service_name} is not among the services created together. Valid services are",
                families,
            )
        if not keypresent("DependsOn", service):
            add_independant_rules(dst_family, service_name, root_stack)
        else:
            src_family = settings.families[service_name]
            if dst_family.stack.title not in src_family.stack.DependsOn:
                src_family.stack.DependsOn.append(dst_family.stack.title)
            dst_family_sg_param = Parameter(
                f"{dst_family.stack.title}GroupId", Type=SG_ID_TYPE
            )
            add_parameters(src_family.template, [dst_family_sg_param])
            src_family.stack.Parameters.update(
                {
                    dst_family_sg_param.title: GetAtt(
                        dst_family.stack.title,
                        f"Outputs.{dst_family.logical_name}GroupId",
                    ),
                }
            )
            for port in dst_family.service_config.network.ports:
                common_args = {
                    "FromPort": port["published"],
                    "ToPort": port["published"],
                    "IpProtocol": port["protocol"],
                    "SourceSecurityGroupOwnerId": Ref(AWS_ACCOUNT_ID),
                    "Description": Sub(
                        f"From ${{{SERVICE_NAME_T}}} to {dst_family.stack.title} on port {port['published']}"
                    ),
                }
                src_family.template.add_resource(
                    SecurityGroupIngress(
                        f"From{src_family.logical_name}To{dst_family.stack.title}On{port['published']}",
                        SourceSecurityGroupId=GetAtt(
                            src_family.ecs_service.sg, "GroupId"
                        ),
                        GroupId=Ref(dst_family_sg_param),
                        **common_args,
                    )
                )


class ServiceNetworking(Ingress):
    """
    Class to group the configuration for Service network settings
    """

    self_key = "Myself"

    def __init__(self, family):
        """
        Initialize network settings for the family ServiceConfig

        :param ecs_composex.common.compose_services.ComposeFamily family:
        """
        self.ports = []
        self.networks = {}
        self.merge_services_ports(family)
        self.merge_networks(family)
        self.configuration = merge_services_network(family)
        self.use_cloudmap = self.configuration["UseCloudmap"]
        self.is_public = self.configuration["IsPublic"]
        self.ingress_from_self = False
        super().__init__(self.configuration[self.master_key], self.ports)
        self.ingress_from_self = keyisset(self.self_key, self.definition)

    def merge_networks(self, family):
        """
        Method to merge network
        """
        for svc in family.services:
            if svc.networks:
                self.networks.update(svc.networks)

    def merge_services_ports(self, family):
        """
        Function to merge two sections of ports

        :param ecs_composex.common.compose_services.ComposeFamily family:
        :return:
        """
        source_ports = [
            service.ports for service in family.ordered_services if service.ports
        ]
        for port_set in source_ports:
            f_source_ports = set_service_ports(self.ports)
            f_override_ports = set_service_ports(port_set)
            self.ports = []
            f_overide_ports_targets = [port["target"] for port in f_override_ports]
            for port in f_override_ports:
                self.ports.append(port)
                for s_port in f_source_ports:
                    if s_port["target"] not in f_overide_ports_targets:
                        self.ports.append(s_port)

    def add_self_ingress(self, family):
        """
        Method to allow communications internally to the group on set ports

        :param ecs_composex.common.compose_services.ComposeFamily family:
        :return:
        """
        if not family.template or not family.ecs_service or not self.ingress_from_self:
            return
        for port in self.ports:
            self.to_self_rules.append(
                SecurityGroupIngress(
                    f"AllowingInterCommunicationPort{port['published']}{port['protocol']}",
                    template=family.template,
                    FromPort=port["published"],
                    ToPort=port["published"],
                    IpProtocol=port["protocol"],
                    GroupId=GetAtt(family.ecs_service.sg, "GroupId"),
                    SourceSecurityGroupId=GetAtt(family.ecs_service.sg, "GroupId"),
                    SourceSecurityGroupOwnerId=Ref(AWS_ACCOUNT_ID),
                    Description=Sub(
                        f"Allowing traffic internally on port {port['published']}"
                    ),
                )
            )

    def add_lb_ingress(self, family, lb_name, lb_sg_ref):
        """
        Method to add ingress rules from other AWS Sources

        :param ecs_composex.common.compose_services.ComposeFamily family:
        :param str lb_name:
        :param lb_sg_ref:
        :return:
        """
        if not family.template or not family.ecs_service:
            return
        for port in self.ports:
            title = f"FromLB{lb_name}To{family.stack.title}On{port['target']}"
            common_args = {
                "FromPort": port["target"],
                "ToPort": port["target"],
                "IpProtocol": port["protocol"],
                "GroupId": GetAtt(family.ecs_service.sg, "GroupId"),
                "SourceSecurityGroupOwnerId": Ref(AWS_ACCOUNT_ID),
                "Description": Sub(
                    f"From ELB {lb_name} to ${{{SERVICE_NAME_T}}} on port {port['target']}"
                ),
            }
            if title in family.template.resources:
                return
            SecurityGroupIngress(
                title,
                template=family.template,
                SourceSecurityGroupId=lb_sg_ref,
                **common_args,
            )



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Functions to format CFN template Outputs
"""

from troposphere import (
    AWS_STACK_NAME,
    AWSHelperFn,
    AWSObject,
    Export,
    If,
    ImportValue,
    Output,
    Sub,
)

from ecs_composex.common import LOG
from ecs_composex.common.cfn_conditions import USE_STACK_NAME_CON_T
from ecs_composex.common.cfn_params import ROOT_STACK_NAME_T, Parameter
from ecs_composex.common.ecs_composex import CFN_EXPORT_DELIMITER


def validate(value):
    """
    Method to validate the input
    :raises: ValueError
    """
    if not len(value) == 3:
        raise ValueError(
            "Output argument expects Name, AttributeName, Value. Only got",
            len(value),
        )
    if not isinstance(value[0], (Parameter, str)):
        raise TypeError("Name should be of type", str, Parameter, "Got", type(value[0]))
    if not isinstance(value[1], str):
        raise TypeError("AttributeName should be of type", str, "Got", type(value[1]))

    valid_type = issubclass(type(value[2]), AWSHelperFn)
    if not (valid_type or isinstance(value[2], (str, int))):
        raise TypeError("Value type is", type(value[2]), "Expected", str, AWSHelperFn)


class ComposeXOutput(object):
    """
    Class to make the output easier.
    """

    delim = CFN_EXPORT_DELIMITER
    stack_string_base = f"${{{AWS_STACK_NAME}}}{delim}"
    root_string_base = f"${{{ROOT_STACK_NAME_T}}}{delim}"

    def __init__(self, resource, values, export=True, duplicate_attr=False):
        """
        Initialize the output class.

        :param resource: The object to export attributes for.
        :param bool export:
        """
        self.object_repr = None
        self.duplicate_attr = duplicate_attr
        self.validate_input(resource, values, export, duplicate_attr)
        self.values = values
        self.outputs = []

        for value in self.values:
            if not isinstance(value, tuple):
                raise TypeError(
                    "All values should be a tuple of (str, value). Got",
                    type(value),
                )
            validate(value)
            attr_name = (
                value[0] if not isinstance(value[0], Parameter) else value[0].title
            )
            output_ext = value[1]
            attr_value = value[2]
            stack_string = (
                f"{self.stack_string_base}{self.object_repr}{self.delim}{attr_name}"
            )
            root_string = (
                f"{self.root_string_base}{self.object_repr}{self.delim}{attr_name}"
            )
            output_name = f"{self.object_repr}{output_ext}"
            output = Output(output_name, Value=attr_value)
            if export:
                output.Export = Export(
                    If(
                        USE_STACK_NAME_CON_T,
                        Sub(stack_string),
                        Sub(root_string),
                    )
                )
            self.outputs.append(output)
            if self.duplicate_attr:
                output = Output(attr_name, Value=attr_value)
                self.outputs.append(output)

    def validate_input(self, resource, values, export=True, duplicate_attr=False):
        if issubclass(type(resource), AWSObject) and hasattr(resource, "title"):
            self.object_repr = resource.title
        elif isinstance(resource, str):
            self.object_repr = resource
        elif isinstance(resource, Parameter):
            self.object_repr = resource.title
        elif resource is None:
            self.object_repr = ""
            if duplicate_attr:
                self.duplicate_attr = False
        else:
            raise TypeError(
                "object should be a subclass of",
                AWSObject,
                "Got",
                type(resource),
            )
        if not isinstance(values, list):
            raise TypeError("values must be of type", list)


def get_import_value(title, attribute_name, delimiter=None):
    """
    Wrapper function to define ImportValue for defined resource name

    :param title: name of the resource exported
    :param attribute_name: attribute exported
    :param delimiter: delimiter between stack name, resource name and attribute
    :return:
    """

    if delimiter is None:
        delimiter = CFN_EXPORT_DELIMITER
    elif not isinstance(delimiter, str):
        LOG.error(
            f"delimiter must be of type str, got {type(delimiter)}. Setting to default"
        )
        delimiter = CFN_EXPORT_DELIMITER
    return ImportValue(
        Sub(f"${{{ROOT_STACK_NAME_T}}}{delimiter}{title}{delimiter}{attribute_name}")
    )



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

from os import path

from ecs_composex.common.cfn_params import Parameter
from ecs_composex.common.ecs_composex import X_KEY

RES_KEY = f"{X_KEY}{path.basename(path.dirname(path.abspath(__file__)))}"

ALARM_NAME_T = "AlarmName"
ALARM_NAME = Parameter(ALARM_NAME_T, Type="String")

ALARM_ARN_T = "AlarmArn"
ALARM_ARN = Parameter(ALARM_ARN_T, return_value="Arn", Type="String")



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to find the SQS streams in lookup
"""

import re

from botocore.exceptions import ClientError
from compose_x_common.compose_x_common import keyisset

from ecs_composex.common import LOG
from ecs_composex.common.aws import (
    define_lookup_role_from_info,
    find_aws_resource_arn_from_tags_api,
)
from ecs_composex.kinesis.kinesis_params import STREAM_ARN, STREAM_KMS_KEY_ID


def get_stream_config(logical_name, stream_arn, session):
    """

    :param str stream_arn:
    :param boto3.session.Session session:
    :return:
    """
    stream_parts = re.compile(
        r"(?:^arn:aws(?:-[a-z]+)?:kinesis:)([\S]+):([0-9]{12}):stream/([\S]+)$"
    )
    stream_name = stream_parts.match(stream_arn).groups()[2]
    client = session.client("kinesis")
    stream_config = {}
    try:
        stream_r = client.describe_stream(StreamName=stream_name)["StreamDescription"]
        stream_config.update(
            {
                logical_name: stream_r["StreamName"],
                STREAM_ARN.title: stream_r["StreamARN"],
            }
        )
        if keyisset("KeyId", stream_r):
            stream_config.update({STREAM_KMS_KEY_ID.title: stream_r["KeyId"]})
        return stream_config
    except client.exceptions.ResourceNotFoundException:
        return None
    except ClientError as error:
        LOG.error(error)
        raise


def lookup_stream_config(logical_name, lookup, session):
    """
    Function to find the DB in AWS account

    :param dict lookup: The Lookup definition for DB
    :param boto3.session.Session session: Boto3 session for clients
    :return:
    """
    kinesis_types = {
        "kinesis": {
            "regexp": r"(?:^arn:aws(?:-[a-z]+)?:kinesis:[\S]+:[0-9]{12}:stream/)([\S]+)$"
        },
    }
    lookup_session = define_lookup_role_from_info(lookup, session)
    stream_arn = find_aws_resource_arn_from_tags_api(
        lookup,
        lookup_session,
        "kinesis",
        types=kinesis_types,
    )
    if not stream_arn:
        return None
    config = get_stream_config(logical_name, stream_arn, lookup_session)
    LOG.debug(config)
    return config



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Helper to generate default parameter group settings from engine name and version

Strip rds internal params to try and fit within 20 param limit
https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-rds-dbparametergroup.html#cfn-rds-dbparametergroup-parameters

"""
import boto3
from botocore.exceptions import ClientError

from ecs_composex.common import LOG


def get_db_cluster_engine_parameter_group_defaults(engine_family):
    """
    Returns a dict of all the parameter group parameters and default values

    :parm str engine_family: Engine family we are getting the cluster settings for, i.e. aurora-mysql5.7
    """

    client = boto3.client("rds")
    try:
        req = client.describe_engine_default_cluster_parameters(
            DBParameterGroupFamily=engine_family
        )
    except ClientError as error:
        LOG.error(error)
        return None
    params_return = {}
    if "EngineDefaults" in req.keys():
        params = req["EngineDefaults"]["Parameters"]
        for param in params:
            if (
                "ParameterValue" in param.keys()
                and "{" not in param["ParameterValue"]
                and "IsModifiable" in param.keys()
                and param["IsModifiable"] is True
                and not param["ParameterName"].startswith("rds.")
            ):
                params_return[param["ParameterName"]] = param["ParameterValue"]
            if param["ParameterName"] == "binlog_format":
                params_return[param["ParameterName"]] = "MIXED"
    return params_return


def get_family_from_engine_version(
    engine_name, engine_version, session=None, client=None
):
    """
    Function to get the engine family from engine name and version
    :param client: override client for boto3 call
    :type client: boto3.client
    :param session: override session for boto3 client
    :type session: boto3.session.Session
    :param engine_name: engine name, ie. aurora-mysql
    :type engine_name: str
    :param engine_version: engine version, ie. 5.7.12
    :type engine_version: str
    :return: engine_family
    :rtype: str
    """
    if not client:
        if not session:
            session = boto3.session.Session()
        client = session.client("rds")
    try:
        req = client.describe_db_engine_versions(
            Engine=engine_name, EngineVersion=engine_version
        )
    except ClientError as error:
        LOG.error(error)
        return None

    db_family = req["DBEngineVersions"][0]["DBParameterGroupFamily"]
    return db_family


def get_family_settings(db_family):
    """
    Function to get the DB family settings
    :param str db_family: The DB family
    :return: db settings or None
    :rtype: None or dict
    """
    if (
        db_family is not None
        and isinstance(db_family, str)
        and db_family.startswith("aurora")
    ):
        LOG.debug("Aurora based instance")
        LOG.debug(f"Looking for parameters for {db_family}")
        return get_db_cluster_engine_parameter_group_defaults(db_family)
    else:
        return None



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to manage Routers specifically.
"""
from compose_x_common.compose_x_common import keyisset
from troposphere import AWS_NO_VALUE, GetAtt, Ref, Sub, appmesh

from ecs_composex.appmesh import appmesh_conditions
from ecs_composex.appmesh.appmesh_params import (
    LISTENER_KEY,
    MATCH_KEY,
    METHOD_KEY,
    NAME_KEY,
    NODES_KEY,
    PORT_KEY,
    PREFIX_KEY,
    PROTOCOL_KEY,
    ROUTES_KEY,
    SCHEME_KEY,
)
from ecs_composex.common import LOG, NONALPHANUM


def define_http_route(route_match, route_nodes):
    route = appmesh.HttpRoute(
        Match=appmesh.HttpRouteMatch(
            Prefix=route_match[PREFIX_KEY]
            if keyisset(PREFIX_KEY, route_match)
            else Ref(AWS_NO_VALUE),
            Scheme=route_match[SCHEME_KEY]
            if keyisset(SCHEME_KEY, route_match)
            else Ref(AWS_NO_VALUE),
            Method=route_match[METHOD_KEY]
            if keyisset(METHOD_KEY, route_match)
            else Ref(AWS_NO_VALUE),
        ),
        Action=appmesh.HttpRouteAction(
            WeightedTargets=[
                appmesh.WeightedTarget(
                    VirtualNode=node.get_node_param,
                    Weight=node.weight,
                )
                for node in route_nodes
            ]
        ),
    )
    return route


def define_route_name(route_match):
    """
    Function to create the route name for an AppMesh Router.

    :param dict route_match: The route argument.
    :return:
    """
    prefix = PREFIX_KEY
    method = METHOD_KEY
    scheme = SCHEME_KEY

    allowed_methods = [
        "GET",
        "CONNECT",
        "DELETE",
        "HEAD",
        "OPTIONS",
        "PATCH",
        "POST",
        "PUT",
        "TRACE",
    ]
    allowed_schemes = ["Http", "Https"]

    prefix_suffix = ""
    method_suffix = ""
    scheme_suffix = ""

    if keyisset(prefix, route_match) and not route_match[prefix].startswith("/"):
        raise ValueError(f"Route {route_match[prefix]} does not start with /")
    elif keyisset(prefix, route_match) and route_match[prefix].startswith("/"):
        prefix_suffix = "".join([key.title() for key in route_match[prefix].split("/")])

    if keyisset(method, route_match) and route_match[method] in allowed_methods:
        method_suffix = route_match[method].title()
    elif keyisset(method, route_match) and route_match[method] not in allowed_methods:
        raise ValueError("Match method must be one of", allowed_methods)
    if keyisset(scheme, route_match) and route_match[scheme] in allowed_schemes:
        method_suffix = route_match[scheme].title()
    elif keyisset(scheme, route_match) and route_match[scheme] not in allowed_schemes:
        raise ValueError("Match scheme must be one of", allowed_schemes)
    return f"{scheme_suffix}{method_suffix}{prefix_suffix}"


class MeshRouter(object):
    """
    Defines a router.
    """

    tcp_routes_keys = [NODES_KEY]
    http_routes_keys = [MATCH_KEY, NODES_KEY]

    def __init__(self, name, definition, mesh, nodes):
        """
        Method to initialize the router

        :param str name:
        :param dict definition:
        :param troposphere.appmesh.Mesh mesh: The mesh to add the router to.
        :param dict nodes: list of nodes defined in the mesh.
        """

        self.title = NONALPHANUM.sub("", name)
        self.definition = definition
        self.validate_definition()
        self.mesh = mesh
        self.port = self.definition[LISTENER_KEY][PORT_KEY]
        self.protocol = self.definition[LISTENER_KEY][PROTOCOL_KEY]
        self.raw_routes = self.definition[ROUTES_KEY]
        self.routes = []
        self.nodes = []
        self.router = appmesh.VirtualRouter(
            f"VirtualRouter{self.title}",
            MeshName=appmesh_conditions.get_mesh_name(mesh),
            MeshOwner=appmesh_conditions.set_mesh_owner_id(),
            VirtualRouterName=Sub(f"{self.title}-vr-${{AWS::StackName}}"),
            Spec=appmesh.VirtualRouterSpec(
                Listeners=[
                    appmesh.VirtualRouterListener(
                        PortMapping=appmesh.PortMapping(
                            Port=self.port, Protocol=self.protocol
                        )
                    )
                ]
            ),
        )
        self.add_routes(nodes)

    def validate_definition(self):
        """
        Method to validate the router definition
        """
        if not keyisset(ROUTES_KEY, self.definition):
            raise KeyError(f"No routes defined for the router {self.title}")
        routes = self.definition[ROUTES_KEY]
        if not keyisset(LISTENER_KEY, self.definition):
            raise KeyError(f"No listener configured for router {self.title}")
        listener = self.definition[LISTENER_KEY]
        if not keyisset(PORT_KEY, listener) or not keyisset(PROTOCOL_KEY, listener):
            raise KeyError("Listener for router requires Port and Protocol")
        if not listener[PROTOCOL_KEY] in routes.keys():
            raise ValueError(
                f"The virtual router is configured for {listener[PROTOCOL_KEY]} but no such route configured"
            )

    def handle_http_route(self, routes, router, nodes, http2=False):
        """
        Function to create a HTTP or HTTP/2 route

        :param list routes: routes of HTTP or HTTP2 protocol
        :param troposphere.appmesh.VirtualRouter router: The virtual router to attach the route to.
        :param dict nodes: list of nodes.
        :param http2: whether it is http2
        :return:
        """
        for route in routes:
            if not all(key in [MATCH_KEY, NODES_KEY] for key in route.keys()):
                raise AttributeError(
                    "Each route must have match and nodes. Got", route.keys()
                )
            route_nodes = []
            for node in route[NODES_KEY]:
                if node[NAME_KEY] in nodes.keys():
                    route_nodes.append(nodes[node[NAME_KEY]])
                else:
                    raise ValueError(
                        f"node {node[NAME_KEY]} is not defined as a virtual node."
                    )
            route_match = route[MATCH_KEY]
            route = define_http_route(route_match, route_nodes)
            self.nodes += [node for node in route_nodes]
            route_name = define_route_name(route_match)
            protocol = "HttpRoute"
            if http2:
                protocol = "Http2Route"
            self.routes.append(
                appmesh.Route(
                    f"{router.title}{route_name}",
                    MeshName=GetAtt(router, "MeshName"),
                    MeshOwner=GetAtt(router, "MeshOwner"),
                    VirtualRouterName=GetAtt(router, "VirtualRouterName"),
                    RouteName=route_name,
                    Spec=appmesh.RouteSpec(**{protocol: route}),
                )
            )

    def handle_tcp_route(self, routes, router, nodes):
        """
        Function to create the TCP routes for the router

        :param list routes: routes of TCP protocol
        :param troposphere.appmesh.VirtualRouter router: The virtual router to attach the route to.
        :param dict nodes: Nodes in the mesh
        """
        for route in routes:
            if not all(key in [NODES_KEY] for key in route.keys()):
                raise AttributeError("Each route must have nodes. Got", route.keys())
            route_nodes = []
            for node in route[NODES_KEY]:
                if node[NAME_KEY] in nodes.keys():
                    route_nodes.append(nodes[node[NAME_KEY]])
                else:
                    raise ValueError(
                        f"node {node[NAME_KEY]} is not defined as a virtual node."
                    )
            route = appmesh.TcpRoute(
                Timeout=appmesh.TcpTimeout(Idle=appmesh.Duration(Unit="ms", Value=1))
                if keyisset("Timeout", route)
                else Ref(AWS_NO_VALUE),
                Action=appmesh.TcpRouteAction(
                    WeightedTargets=[
                        appmesh.WeightedTarget(
                            VirtualNode=node.get_node_param,
                            Weight=node.weight,
                        )
                        for node in route_nodes
                    ]
                ),
            )
            protocol = "TcpRoute"
            self.routes.append(
                appmesh.Route(
                    f"{router.title}{protocol}",
                    MeshName=GetAtt(router, "MeshName"),
                    MeshOwner=GetAtt(router, "MeshOwner"),
                    VirtualRouterName=GetAtt(router, "VirtualRouterName"),
                    RouteName=Sub(f"${{AWS::StackName}}{protocol}"),
                    Spec=appmesh.RouteSpec(**{protocol: route}),
                )
            )

    def add_routes(self, nodes):
        """
        Method to register routers
        """
        for route_protocol in self.raw_routes.keys():
            if route_protocol != self.protocol:
                raise ValueError(
                    f"he virtual router is configured for {self.protocol} "
                    f"but a route for protocol {route_protocol} has been found. This is not supported."
                )

            if route_protocol == "Http" or route_protocol == "Http2":
                self.handle_http_route(
                    self.raw_routes[route_protocol],
                    self.router,
                    nodes,
                    eval('route_protocol == "Http2"'),
                )
            elif route_protocol == "Tcp":
                self.handle_tcp_route(
                    self.raw_routes[route_protocol], self.router, nodes
                )
            elif route_protocol == "gRPC":
                LOG.warning("gRPC is not yet supported. Sorry.")



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Main module. Implements all the logic of the macro event parsing and generating.
Passes all the arguments to ECS ComposeX to render the CFN templates for the docker-compose file.
"""

import re
import tempfile
from copy import deepcopy
from os import environ, path
from urllib.parse import urlparse

import boto3
import requests
import yaml
from compose_x_common.compose_x_common import keyisset

from ecs_composex.common import LOG
from ecs_composex.common.settings import ComposeXSettings
from ecs_composex.common.stacks import process_stacks
from ecs_composex.ecs_composex import generate_full_template


def set_settings_from_remote_files(files, settings_params, folder, session=None):
    if not keyisset(ComposeXSettings.input_file_arg, settings_params):
        local_files = []
        settings_params[ComposeXSettings.input_file_arg] = local_files
    else:
        local_files = settings_params[ComposeXSettings.input_file_arg]
    if not session:
        session = boto3.session.Session()
    client = session.client("s3")
    bucket_re = re.compile(r"(?:s3://)([a-z0-9-.]+)/([\S]+$)")
    for file in files:
        if file.startswith("s3://") and bucket_re.match(file):
            try:
                bucket_name = bucket_re.match(file).groups()[0]
                file_path = bucket_re.match(file).groups()[1]
                file_name = path.basename(file_path)
                new_file_path = path.abspath(f"{folder.name}/{file_name}")
                with open(new_file_path, "wb") as fd:
                    file_r = client.get_object(Bucket=bucket_name, Key=file_path)
                    fd.write(file_r["Body"].read())
                    local_files.append(new_file_path)
            except client.exceptions.NoSuchKey:
                LOG.error(f"Failed to download the file {file} from S3")
                raise
        elif file.startswith("https://"):
            file_name = path.basename(urlparse(file).path)
            new_file_path = path.abspath(f"{folder.name}/{file_name}")
            file_r = requests.get(file)
            if not file_r.status_code == "200":
                raise FileNotFoundError(f"Unable to retrieve {file}")
            with open(new_file_path, "w") as fd:
                fd.write(file_r.text)
            local_files.append(new_file_path)


def settings_from_raw_content(settings, content, folder):
    """
    Function to define the ComposeX Settings from RAW Content
    """
    file_path = f"{folder.name}/src.yml"
    with open(file_path, "w") as fd:
        fd.write(yaml.dump(content))
    if not keyisset(ComposeXSettings.input_file_arg, settings):
        settings.update({ComposeXSettings.input_file_arg: [file_path]})


def init_settings_params(settings_params, fragment, request_id, folder):
    """
    Function to define the parameters to send to ECS ComposeX Settings

    :param dict settings_params:
    :param fragment:
    :param str request_id:
    :param folder: Temporary folder to store all the files into.
    :return:
    """
    new_fragment = {}
    settings_params.update({"command": "create"})
    if keyisset("version", fragment) or keyisset("services", fragment):
        settings_from_raw_content(settings_params, fragment, folder)
    elif keyisset("Raw", settings_params):
        settings_from_raw_content(settings_params, settings_params["Raw"], folder)
    if keyisset("ComposeFiles", settings_params):
        set_settings_from_remote_files(
            settings_params["ComposeFiles"], settings_params, folder
        )
    if not keyisset("Name", settings_params):
        settings_params.update({"Name": request_id})
    return new_fragment


def define_s3_bucket_upload(settings, params):
    """
    Function to override the buckets settings before rendering
    Priority goes to BucketName if defined in the CFN template in the macro Parameters
    Fall back to env var UPLOAD_BUCKET_NAME if set, else, sticks to default behaviour
    """
    if keyisset(settings.bucket_arg, params):
        settings.bucket_name = params["BucketName"]
        LOG.info(f"Overriding bucket name to {settings.bucket_name}")
    elif environ.get("UPLOAD_BUCKET_NAME"):
        settings.bucket_name = environ.get("UPLOAD_BUCKET_NAME")
        LOG.info(
            f"Found BUCKET_NAME defined in the environment variables. Setting to {settings.bucket_name}"
        )


def lambda_handler(event, context):
    """
    Lambda function entrypoint.
    """
    response = {"status": "success", "requestId": event["requestId"]}
    region = event["region"]
    account_id = event["accountId"]
    transform_id = event["transformId"]
    LOG.info(
        f"Processing in {region} for transform {transform_id} in account {account_id}"
    )

    params = event["params"]
    fragment = event["fragment"]
    request_id = event["requestId"]

    folder = tempfile.TemporaryDirectory(prefix=request_id)

    settings_params = deepcopy(params)

    new_fragment = init_settings_params(settings_params, fragment, request_id, folder)
    settings = ComposeXSettings(for_macro=True, **settings_params)
    settings.set_bucket_name_from_account_id()
    define_s3_bucket_upload(settings, params)
    settings.set_azs_from_api()
    settings.deploy = True
    settings.upload = True

    root_stack = generate_full_template(settings)
    process_stacks(root_stack, settings)
    rendered_template = root_stack.stack_template.to_dict()
    for key in [
        "Resources",
        "Outputs",
        "Mappings",
        "Conditions",
        "Metadata",
    ]:
        new_fragment[key] = (
            rendered_template[key] if keyisset(key, rendered_template) else {}
        )
    if "Parameters" in new_fragment.keys():
        del new_fragment["Parameters"]
    response["fragment"] = new_fragment
    folder.cleanup()
    return response



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to import CFN Resources defined by their properties
"""

from inspect import isfunction

from compose_x_common.compose_x_common import keyisset, keypresent
from troposphere import AWSHelperFn, AWSProperty


def handle_list(properties, property_class):
    """
    Function to handle list properties

    :param property_class:
    :param properties:
    :return:
    """
    rendered_properties = []
    for property_definition in properties:
        if not isinstance(property_definition, (str, int, float, bool)):
            record = import_record_properties(property_definition, property_class)
            rendered_properties.append(property_class(**record))
        else:
            rendered_properties.append(property_definition)
    return rendered_properties


def import_non_functions(props, prop_name, top_class, properties, set_to_novalue):
    """
    Function to set property for flat object or recursive to sub properties

    :param dict props:
    :param str prop_name:
    :param top_class:
    :param dict properties:
    :param bool set_to_novalue:
    """
    if isinstance(properties[prop_name], AWSHelperFn):
        props[prop_name] = properties[prop_name]
    elif isinstance(properties[prop_name], (str, int, float, tuple)) or top_class.props[
        prop_name
    ][0] in (
        str,
        int,
        float,
    ):
        if top_class.props[prop_name][0] in (str, int, float):
            props[prop_name] = top_class.props[prop_name][0](properties[prop_name])
        else:
            props[prop_name] = properties[prop_name]
    elif isinstance(properties[prop_name], dict):
        try:
            if issubclass(top_class.props[prop_name][0], AWSProperty):
                sub_props = import_record_properties(
                    properties[prop_name],
                    top_class.props[prop_name][0],
                    set_to_novalue,
                    ignore_missing_required=False,
                )
                props[prop_name] = top_class.props[prop_name][0](**sub_props)
            else:
                props[prop_name] = properties[prop_name]
        except TypeError:
            props[prop_name] = properties[prop_name]


def import_record_properties(
    properties, top_class, set_to_novalue=False, ignore_missing_required=True
):
    """
    Generic function importing the RecordSet properties.
    If the property was not defined, it is either left empty or set to AWS::NoValue
    For inner recursive, we enforce check on required properties.

    :param dict properties:
    :param top_class: The class we are going to import properties for
    :param bool set_to_novalue: Instead of skipping the property, actively set to AWS::NoValue
    :param bool ignore_missing_required: Whether or not raise an error when missing an essential key.
    :return:  The properties for the RecordSet
    :rtype: dict
    """
    props = {}
    for prop_name in top_class.props:
        if not keypresent(prop_name, properties) and not top_class.props[prop_name][1]:
            continue
        elif (
            not keypresent(prop_name, properties)
            and top_class.props[prop_name][1]
            and not ignore_missing_required
        ):
            raise KeyError(
                f"Property {prop_name} is required for the definition of {top_class}"
            )
        elif keyisset(prop_name, properties) and isinstance(
            top_class.props[prop_name][0], list
        ):
            props[prop_name] = handle_list(
                properties[prop_name], top_class.props[prop_name][0][0]
            )
        elif keypresent(prop_name, properties) and isfunction(
            top_class.props[prop_name][0]
        ):
            props[prop_name] = properties[prop_name]
        elif keypresent(prop_name, properties) and not isfunction(
            properties[prop_name]
        ):
            import_non_functions(
                props, prop_name, top_class, properties, set_to_novalue
            )
        elif keypresent(prop_name, properties):
            props[prop_name] = properties[prop_name]
    return props



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""
Module to find the SQS queues in lookup
"""

import re

from ecs_composex.common import LOG
from ecs_composex.common.aws import (
    define_lookup_role_from_info,
    find_aws_resource_arn_from_tags_api,
)
from ecs_composex.ssm_parameter.ssm_params import SSM_PARAM_ARN, SSM_PARAM_NAME


def lookup_param_config(lookup, session):
    """
    Function to find the DB in AWS account

    :param dict lookup: The Lookup definition for DB
    :param boto3.session.Session session: Boto3 session for clients
    :return:
    """
    ssm_re = re.compile(r"(?:^arn:aws(?:-[a-z]+)?:ssm:[\S]+:[0-9]+:parameter)(/[\S]+)$")
    ssm_types = {
        "ssm": {"regexp": ssm_re.pattern},
    }
    lookup_session = define_lookup_role_from_info(lookup, session)
    param_arn = find_aws_resource_arn_from_tags_api(
        lookup,
        lookup_session,
        "ssm",
        types=ssm_types,
    )
    if not param_arn:
        return None
    config = {
        SSM_PARAM_NAME.title: ssm_re.match(param_arn).groups()[-1],
        SSM_PARAM_ARN.title: param_arn,
    }
    LOG.debug(config)
    return config
